import unittest.mock
from Bio import SeqIO, Seq
from ..io import fasta
from ..orm import cv, organism, sequence


class TestFastaImport(unittest.TestCase):
    """Tests various functions used to load a FASTA file into a database"""

    @classmethod
    def setUpClass(cls):
        """Creates an instance of the base class to be tested and instantiates global attributes"""
        cls.client = fasta.FastaImportClient("testuri", test_environment=True)
        cls.client._sequence_terms = {
            "contig": cv.CvTerm(cv_id=4, dbxref_id=41, name="contig", cvterm_id=41),
            "supercontig": cv.CvTerm(cv_id=4, dbxref_id=42, name="supercontig", cvterm_id=42),
            "chromosome": cv.CvTerm(cv_id=4, dbxref_id=43, name="chromosome", cvterm_id=43),
            "region": cv.CvTerm(cv_id=4, dbxref_id=44, name="region", cvterm_id=44)
        }
        cls.client._top_level_term = cv.CvTerm(cv_id=11, dbxref_id=91, name="top_level_seq", cvterm_id=91)

    def setUp(self):
        # Create a default FASTA record
        self.default_fasta_record = SeqIO.SeqRecord(seq="ACTGAAC", id="testid", name="testname",
                                                    description="sequence_name | organism=testorganism | SO=chromosome")

    @unittest.mock.patch("pychado.io.fasta.FastaImportClient._handle_feature")
    @unittest.mock.patch("pychado.io.fasta.FastaImportClient._create_feature")
    @unittest.mock.patch("pychado.io.fasta.FastaImportClient._extract_type")
    def test_handle_sequence(self, mock_extract_type: unittest.mock.Mock, mock_create: unittest.mock.Mock,
                             mock_insert: unittest.mock.Mock):
        # Tests the function transferring data from a FASTA record to the 'feature' table
        self.assertIs(mock_extract_type, self.client._extract_type)
        self.assertIs(mock_create, self.client._create_feature)
        self.assertIs(mock_insert, self.client._handle_feature)

        organism_entry = organism.Organism(genus="", species="", abbreviation="testorganism", organism_id=33)
        type_entry = cv.CvTerm(name="contig", cv_id=1, dbxref_id=2, cvterm_id=77)

        mock_extract_type.return_value = None
        self.client._handle_sequence(self.default_fasta_record, organism_entry, type_entry)
        mock_extract_type.assert_called_with(self.default_fasta_record)
        mock_create.assert_called_with(self.default_fasta_record, 33, 77)
        mock_insert.assert_called()

        mock_extract_type.return_value = "chromosome"
        self.client._handle_sequence(self.default_fasta_record, organism_entry, type_entry)
        mock_extract_type.assert_called_with(self.default_fasta_record)
        mock_create.assert_called_with(self.default_fasta_record, 33, 43)
        mock_insert.assert_called()

    @unittest.mock.patch("pychado.io.fasta.FastaImportClient._handle_featureprop")
    @unittest.mock.patch("pychado.orm.sequence.FeatureProp")
    @unittest.mock.patch("pychado.io.fasta.FastaImportClient.query_all")
    def test_mark_as_top_level_sequence(self, mock_query: unittest.mock.Mock, mock_featureprop: unittest.mock.Mock,
                                        mock_insert: unittest.mock.Mock):
        # Tests the function transferring data from a FASTA record to the 'feature' table
        self.assertIs(mock_query, self.client.query_all)
        self.assertIs(mock_featureprop, sequence.FeatureProp)
        self.assertIs(mock_insert, self.client._handle_featureprop)
        feature_entry = sequence.Feature(organism_id=1, type_id=2, uniquename="", feature_id=33)

        self.client._mark_as_top_level_sequence(feature_entry)
        mock_query.assert_called_with(sequence.FeatureProp, feature_id=33)
        mock_featureprop.assert_called_with(feature_id=33, type_id=91, value="true")
        mock_insert.assert_called()

    def test_create_feature(self):
        # Tests the function creating an entry for the 'feature' table from a FASTA record
        feature_entry = self.client._create_feature(self.default_fasta_record, 1, 2)
        self.assertEqual(feature_entry.organism_id, 1)
        self.assertEqual(feature_entry.type_id, 2)
        self.assertEqual(feature_entry.uniquename, "testid")
        self.assertEqual(feature_entry.residues, "ACTGAAC")
        self.assertEqual(feature_entry.seqlen, 7)

    def test_extract_type(self):
        # Tests the function extracting the sequence type from a FASTA record
        sequence_type = self.client._extract_type(self.default_fasta_record)
        self.assertEqual(sequence_type, "chromosome")
        self.default_fasta_record.description = ""
        sequence_type = self.client._extract_type(self.default_fasta_record)
        self.assertIsNone(sequence_type)


class TestFastaExport(unittest.TestCase):
    """Tests various functions used to export a FASTA file from a database"""

    @classmethod
    def setUpClass(cls):
        """Creates an instance of the base class to be tested and instantiates global attributes"""
        cls.client = fasta.FastaExportClient("testuri", test_environment=True)
        cls.client._sequence_terms = {
            "gene": cv.CvTerm(cv_id=4, dbxref_id=41, name="gene", cvterm_id=41),
            "pseudogene": cv.CvTerm(cv_id=4, dbxref_id=42, name="pseudogene", cvterm_id=42),
            "polypeptide": cv.CvTerm(cv_id=4, dbxref_id=43, name="polypeptide", cvterm_id=43)
        }
        cls.client._part_of_term = cv.CvTerm(cv_id=11, dbxref_id=91, name="part_of", is_relationshiptype=1,
                                             cvterm_id=91)
        cls.client._derives_from_term = cv.CvTerm(cv_id=11, dbxref_id=92, name="derives_from",
                                                  is_relationshiptype=1, cvterm_id=92)
        cls.client._top_level_term = cv.CvTerm(cv_id=11, dbxref_id=91, name="top_level_seq", cvterm_id=91)

    @unittest.mock.patch("pychado.io.fasta.FastaExportClient._extract_nucleotide_sequence")
    def test_extract_residues_by_type(self, mock_extract_nucleotides: unittest.mock.Mock):
        # Tests the function that extracts the sequence of nucleotides/amino acids of a feature
        self.assertIs(mock_extract_nucleotides, self.client._extract_nucleotide_sequence)
        feature_entry = sequence.Feature(organism_id=1, type_id=2, uniquename="test", residues="CTGA", feature_id=33)

        mock_extract_nucleotides.return_value = "AAGG"
        residues = self.client._extract_residues_by_type(feature_entry, [], "genes")
        mock_extract_nucleotides.assert_called_with(feature_entry, [])
        self.assertEqual(residues, "AAGG")

        mock_extract_nucleotides.reset_mock()
        residues = self.client._extract_residues_by_type(feature_entry, [], "contigs")
        mock_extract_nucleotides.assert_not_called()
        self.assertEqual(residues, "CTGA")

    def test_are_residues_valid(self):
        # Tests the function that checks if a sequence of nucleotides/amino acids is composed of valid IUPAC codes
        valid = self.client._are_residues_valid("", "genes")
        self.assertFalse(valid)
        valid = self.client._are_residues_valid("agct", "genes")
        self.assertTrue(valid)
        valid = self.client._are_residues_valid("AGCTXX", "genes")
        self.assertFalse(valid)
        valid = self.client._are_residues_valid("MRAB*", "proteins")
        self.assertTrue(valid)
        valid = self.client._are_residues_valid("RMAB*", "proteins")
        self.assertFalse(valid)
        valid = self.client._are_residues_valid("MR*AB*", "proteins")
        self.assertFalse(valid)

    @unittest.mock.patch("pychado.io.fasta.FastaExportClient.query_first")
    def test_extract_nucleotide_sequences(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the nucleotide sequence of a feature from the database
        self.assertIs(mock_query, self.client.query_first)
        feature_entry = sequence.Feature(organism_id=1, type_id=2, uniquename="test", residues="CTGA", feature_id=33)

        mock_query.return_value = sequence.FeatureLoc(feature_id=33, srcfeature_id=34, fmin=1, fmax=3, strand=1)
        residues = self.client._extract_nucleotide_sequence(feature_entry, [])
        mock_query.assert_called_with(sequence.FeatureLoc, feature_id=33)
        self.assertIsNone(residues)

        srcfeature_entries = [sequence.Feature(organism_id=1, type_id=2, uniquename="test", residues="ACTGGTAA",
                                               feature_id=34)]
        residues = self.client._extract_nucleotide_sequence(feature_entry, srcfeature_entries)
        self.assertEqual(residues, "CT")

        mock_query.return_value = sequence.FeatureLoc(feature_id=33, srcfeature_id=34, fmin=0, fmax=6, strand=-1)
        residues = self.client._extract_nucleotide_sequence(feature_entry, srcfeature_entries)
        self.assertEqual(residues, "ACCAGT")

        mock_query.return_value = sequence.FeatureLoc(feature_id=33, srcfeature_id=34, fmin=1, fmax=300)
        residues = self.client._extract_nucleotide_sequence(feature_entry, srcfeature_entries)
        self.assertIsNone(residues)

    @unittest.mock.patch("Bio.SeqIO.SeqRecord")
    @unittest.mock.patch("Bio.Seq.Seq")
    @unittest.mock.patch("pychado.io.fasta.FastaExportClient._create_fasta_attributes")
    def test_create_fasta_record(self, mock_attributes: unittest.mock.Mock, mock_sequence: unittest.mock.Mock,
                                 mock_record: unittest.mock.Mock):
        # Tests the function that creates a FASTA record
        self.assertIs(mock_attributes, self.client._create_fasta_attributes)
        self.assertIs(mock_sequence, Seq.Seq)
        self.assertIs(mock_record, SeqIO.SeqRecord)

        feature_entry = sequence.Feature(organism_id=1, type_id=2, uniquename="test", feature_id=33)
        organism_entry = organism.Organism(genus="testgenus", species="testspecies", organism_id=44)
        type_entry = cv.CvTerm(name="contig", cv_id=1, dbxref_id=2, cvterm_id=55)
        mock_attributes.return_value = "desc"
        mock_sequence.return_value = "seq"

        self.client._create_fasta_record(feature_entry, organism_entry, type_entry, "AGCT", "testrelease", True)
        mock_attributes.assert_called_with(organism_entry, type_entry, "testrelease", True)
        mock_sequence.assert_called_with("AGCT")
        mock_record.assert_called_with("seq", id="test", name="test", description="desc")

    @unittest.mock.patch("pychado.io.fasta.FastaExportClient.query_features_by_property_type")
    def test_extract_srcfeatures_by_type(self, mock_query: unittest.mock.Mock):
        # Tests that the feature table is only queried for certain feature types
        self.assertIs(mock_query, self.client.query_features_by_property_type)

        organism_entry = organism.Organism(genus="testgenus", species="testspecies", organism_id=44)
        self.client._extract_srcfeatures_by_type(organism_entry, "contigs")
        mock_query.assert_not_called()

        self.client._extract_srcfeatures_by_type(organism_entry, "genes")
        mock_query.assert_called_with(44, 91)

    @unittest.mock.patch("pychado.io.fasta.FastaExportClient.query_protein_features")
    @unittest.mock.patch("pychado.io.fasta.FastaExportClient.query_features_by_property_type")
    @unittest.mock.patch("pychado.io.fasta.FastaExportClient.query_features_by_type")
    def test_extract_features_by_type(self, mock_query_genes: unittest.mock.Mock,
                                      mock_query_contigs: unittest.mock.Mock, mock_query_proteins: unittest.mock.Mock):
        # Tests that the feature table is correctly queried depending on the type of features of interest
        self.assertIs(mock_query_genes, self.client.query_features_by_type)
        self.assertIs(mock_query_contigs, self.client.query_features_by_property_type)
        self.assertIs(mock_query_proteins, self.client.query_protein_features)

        organism_entry = organism.Organism(genus="testgenus", species="testspecies", organism_id=44)
        self.client._extract_features_by_type(organism_entry, "contigs")
        mock_query_genes.assert_not_called()
        mock_query_contigs.assert_called_with(44, 91)
        mock_query_proteins.assert_not_called()

        mock_query_genes.reset_mock()
        mock_query_contigs.reset_mock()
        mock_query_proteins.reset_mock()
        self.client._extract_features_by_type(organism_entry, "proteins")
        mock_query_genes.assert_not_called()
        mock_query_contigs.assert_not_called()
        mock_query_proteins.assert_called_with(44, 41, 91, 92)

        mock_query_genes.reset_mock()
        mock_query_contigs.reset_mock()
        mock_query_proteins.reset_mock()
        self.client._extract_features_by_type(organism_entry, "genes")
        mock_query_genes.assert_called_with(44, [41])
        mock_query_contigs.assert_not_called()
        mock_query_proteins.assert_not_called()

    @unittest.mock.patch("pychado.io.fasta.FastaExportClient._release_key_value_pair")
    @unittest.mock.patch("pychado.io.fasta.FastaExportClient._genome_version_key_value_pair")
    @unittest.mock.patch("pychado.io.fasta.FastaExportClient._type_key_value_pair")
    @unittest.mock.patch("pychado.io.fasta.FastaExportClient._organism_key_value_pair")
    def test_create_fasta_attributes(self, mock_organism: unittest.mock.Mock, mock_type: unittest.mock.Mock,
                                     mock_version: unittest.mock.Mock, mock_release: unittest.mock.Mock):
        # Tests the correct creation of a header for a FASTA sequence
        self.assertIs(mock_organism, self.client._organism_key_value_pair)
        self.assertIs(mock_type, self.client._type_key_value_pair)
        self.assertIs(mock_version, self.client._genome_version_key_value_pair)
        self.assertIs(mock_release, self.client._release_key_value_pair)

        organism_entry = organism.Organism(genus="testgenus", species="testspecies", organism_id=33)
        type_entry = cv.CvTerm(name="contig", cv_id=1, dbxref_id=2, cvterm_id=44)
        mock_organism.return_value = "orgname"
        mock_type.return_value = "typename"
        mock_version.return_value = "versionnumber"
        mock_release.return_value = "relname"

        attributes = self.client._create_fasta_attributes(organism_entry, type_entry, "", True)
        mock_organism.assert_called_with(organism_entry)
        mock_type.assert_called_with(type_entry)
        mock_version.assert_not_called()
        mock_release.assert_not_called()
        self.assertEqual(attributes, "| orgname | typename")

        organism_entry.version = 11
        attributes = self.client._create_fasta_attributes(organism_entry, type_entry, "testrelease", True)
        mock_version.assert_called_with(organism_entry)
        mock_release.assert_called_with("testrelease")
        self.assertEqual(attributes, "| orgname | typename | versionnumber | relname")

    def test_organism_key_value_pair(self):
        # Tests the correct creation of a key-value pair for an organism name with proper escaping applied
        organism_entry = organism.Organism(genus="testgenus", species="testspecies", organism_id=33)
        pair = self.client._organism_key_value_pair(organism_entry)
        self.assertEqual(pair, "organism=testgenus%20testspecies")
        organism_entry.infraspecific_name = "teststrain"
        pair = self.client._organism_key_value_pair(organism_entry)
        self.assertEqual(pair, "organism=testgenus%20testspecies%20teststrain")

    def test_genome_version_key_value_pair(self):
        # Tests the correct creation of a key-value pair for a genome version with proper escaping applied
        organism_entry = organism.Organism(genus="testgenus", species="testspecies", organism_id=33, version=11)
        pair = self.client._genome_version_key_value_pair(organism_entry)
        self.assertEqual(pair, "genome_version=11")

    def test_type_key_value_pair(self):
        # Tests the correct creation of a key-value pair for a feature type with proper escaping applied
        type_entry = cv.CvTerm(name="contig", cv_id=1, dbxref_id=2, cvterm_id=44)
        pair = self.client._type_key_value_pair(type_entry)
        self.assertEqual(pair, "sequence_type=contig")
        type_entry.name = "con tig"
        pair = self.client._type_key_value_pair(type_entry)
        self.assertEqual(pair, "sequence_type=con%20tig")

    def test_release_key_value_pair(self):
        # Tests the correct creation of a key-value pair for a release name with proper escaping applied
        pair = self.client._release_key_value_pair("2019-01")
        self.assertEqual(pair, "release=2019-01")
        pair = self.client._release_key_value_pair("2019:01")
        self.assertEqual(pair, "release=2019%3A01")



import os
import tempfile
from typing import List, Dict, Union
import gffutils
from . import iobase, ontology, fasta
from .. import utils
from ..orm import general, cv, organism, pub, sequence


class GFFClient(object):
    """Helper class for GFF-related operations"""

    @staticmethod
    def convert_strand(strand: str) -> Union[None, int]:
        """Converts the strand from string notation to integer notation"""
        if strand == '+':
            return 1
        elif strand == '-':
            return -1
        else:
            return None

    @staticmethod
    def back_convert_strand(strand: Union[None, int]) -> str:
        """Converts the strand from integer notation to string notation"""
        if strand is None:
            return "."
        elif strand > 0:
            return "+"
        elif strand < 0:
            return "-"
        else:
            return "."

    @staticmethod
    def convert_frame(frame: str) -> Union[None, int]:
        """Converts the frame from string notation to integer notation"""
        try:
            phase = int(frame)
        except ValueError:
            return None
        if phase in [0, 1, 2]:
            return phase
        else:
            return None

    @staticmethod
    def back_convert_frame(phase: Union[None, int]) -> str:
        """Converts the frame from integer notation to string notation"""
        if phase is None:
            return "."
        else:
            return str(phase)

    @staticmethod
    def _transcript_types() -> List[str]:
        """Lists considered transcript types"""
        return ["mrna", "rrna", "trna", "snrna", "ncrna", "scrna", "snorna", "pseudogenic_transcript"]

    @staticmethod
    def _feature_relationship_types() -> List[str]:
        """Lists considered feature_relationship types"""
        return ["part_of", "derives_from"]

    @staticmethod
    def _feature_property_types() -> List[str]:
        """Lists considered featureprop types"""
        return ["comment", "description", "score", "source"]

    @staticmethod
    def _synonym_types() -> List[str]:
        """Lists considered synonym types"""
        return ["alias", "synonym", "previous_systematic_id"]


class GFFImportClient(iobase.ChadoClient, GFFClient):
    """Class for importing genomic data from GFF files into Chado"""

    def __init__(self, uri: str, verbose=False, test_environment=False):
        """Constructor"""

        # Connect to database
        self.uri = uri
        self.verbose = verbose
        self.test_environment = test_environment
        if self.test_environment:
            self.printer = utils.VerbosePrinter(self.verbose)
        else:
            super().__init__(self.uri, self.verbose)

        # Create array for temporary SQLite databases
        self._sqlite_databases = []

        # Load essential database entries
        if not self.test_environment:
            self._load_essentials()

    def __del__(self):
        """Destructor"""

        # Disconnect from database
        if not self.test_environment:
            super().__del__()

        # Delete temporary SQLite database
        for database in self._sqlite_databases:
            if os.path.exists(database):
                os.remove(database)

    def _load_essentials(self) -> None:
        """Loads essential database entries"""

        self._part_of_term = self._load_cvterm_from_cv("part_of", "relationship")
        self._derives_from_term = self._load_cvterm_from_cv("derives_from", "sequence")
        self._parent_terms = {self._part_of_term.name: self._part_of_term,
                              self._derives_from_term.name: self._derives_from_term}
        self._parent_type_ids = [self._part_of_term.cvterm_id, self._derives_from_term.cvterm_id]

        self._synonym_terms = self._load_terms_from_cv_dict(
            "genedb_synonym_type", self._synonym_types())
        self._synonym_type_ids = self._extract_cvterm_ids_from_dict(
            self._synonym_terms, self._synonym_types())

        self._feature_property_terms = self._load_terms_from_cv_dict(
            "feature_property", self._feature_property_types())
        self._feature_property_type_ids = self._extract_cvterm_ids_from_dict(
            self._feature_property_terms, self._feature_property_types())

        self._sequence_terms = self._load_terms_from_cv_dict(
            "sequence", ["gene", "intron", "exon", "CDS", "mRNA", "chromosome"])
        self._default_pub = self._load_pub("null")

        self._go_db = self._load_db("GO")

    def load(self, filename: str, organism_name: str, fasta_filename: str, sequence_type: str, fresh_load=False,
             force_purge=False, full_genome=False):
        """Import data from a GFF3 file into a Chado database"""

        # Check for file existence
        if not os.path.exists(filename):
            raise iobase.InputFileError("Input file '" + filename + "' does not exist.")

        # Remove existing database entries, if applicable
        default_organism = self._load_organism(organism_name)
        self._handle_existing_features(default_organism, fresh_load, force_purge)

        # Import FASTA sequences, if present
        self._import_fasta(filename, fasta_filename, organism_name, sequence_type)

        # Create temporary SQLite database
        gff_db = self._create_sqlite_db(filename)

        # Initiate global containers
        all_feature_entries = {}

        # Loop over all entries in the gff file
        for gff_record in gff_db.all_features():

            # Insert, update or delete entries in various tables
            self._insert_gff_record_into_database(gff_record, default_organism, all_feature_entries)

        # Mark obsolete features
        if full_genome and not fresh_load:
            top_level_entries = self._extract_gff_sequence_names(gff_db)
            self._mark_obsolete_features(default_organism, all_feature_entries, top_level_entries)

        # Commit changes
        self.session.commit()

    def _import_fasta(self, gff_file: str, fasta_file: str, organism_name: str, sequence_type: str) -> None:
        """Imports sequences from a FASTA file into the Chado database"""

        # Check if the GFF file contains FASTA sequences
        fasta_is_temporary = False
        if self._has_fasta(gff_file):
            if fasta_file:

                # Error message - only one file with FASTA is permitted
                raise iobase.InputFileError("You cannot provide a GFF file with FASTA sequences "
                                            "plus a separate FASTA file.")
            else:

                # Create a FASTA file containing the sequences in the GFF file
                (fasta_handle, fasta_file) = tempfile.mkstemp(dir=os.getcwd())
                os.close(fasta_handle)
                fasta_is_temporary = True
                self._split_off_fasta(gff_file, fasta_file)

        # Import sequences from FASTA file, if present
        if fasta_file:
            fasta_client = fasta.FastaImportClient(self.uri, self.verbose)
            fasta_client.load(fasta_file, organism_name, sequence_type)

        # Delete temporary file
        if fasta_is_temporary and os.path.exists(fasta_file):
            os.remove(fasta_file)

    @staticmethod
    def _has_fasta(gff_file: str) -> bool:
        """Checks if a GFF file contains a FASTA section"""
        infile = utils.open_file_read(gff_file)
        fasta_started = False
        for line in infile:
            if "##FASTA" in line:
                fasta_started = True
                break
        utils.close(infile)
        return fasta_started

    @staticmethod
    def _split_off_fasta(gff_file: str, fasta_file: str) -> None:
        """Copies the FASTA section from a GFF file into a separate FASTA file"""
        infile = utils.open_file_read(gff_file)
        outfile = utils.open_file_write(fasta_file)
        fasta_started = False
        for line in infile:
            if fasta_started:
                outfile.write(line)
            else:
                if "##FASTA" in line:
                    fasta_started = True
        utils.close(infile)
        utils.close(outfile)

    def _create_sqlite_db(self, filename: str) -> gffutils.FeatureDB:
        """Creates a local SQLite database containing data from a GFF file"""
        database_name = os.path.abspath(filename) + ".sqlitedb"
        self._sqlite_databases.append(database_name)
        gffutils.create_db(filename, database_name, force=True, keep_order=True)
        database = gffutils.FeatureDB(database_name)
        return database

    def _handle_existing_features(self, organism_entry: organism.Organism, fresh_load: bool, force_purge: bool) -> None:
        """Checks if there are existing features for the organism, and deletes them if required"""

        # Check if this is an import "from scratch". If not, there is no need to take any action
        if not fresh_load:
            return

        # Check if the database already contains features for this organism. If not, there is no need to take any action
        existing_features_query = self.query_table(sequence.Feature, organism_id=organism_entry.organism_id)
        if existing_features_query.count() == 0:
            return

        # Check if the '--force' option was used
        if not force_purge:

            # Abort the operation
            raise iobase.DatabaseError("The database already contains features for organism '"
                                       + organism_entry.abbreviation + "'. Run without option '--fresh_load' to update"
                                       + " the features or with option '--force' to overwrite everything.")
        else:

            # Delete all existing features for this organism
            self.printer.print("Deleting all features for organism '" + organism_entry.abbreviation + "'")
            existing_features_query.delete()

    def _mark_obsolete_features(self, organism_entry: organism.Organism,
                                all_features: Dict[str, sequence.Feature], top_level_features: List[str]) -> None:
        """Marks features as obsolete"""

        # Loop over all features for the given organism in the database
        for feature_name in self._load_feature_names(organism_entry):

            # Check if the feature is also present in the input file
            if feature_name not in all_features and feature_name not in top_level_features:

                # Mark the feature as obsolete, if necessary
                self._mark_feature_as_obsolete(organism_entry, feature_name)

    def _insert_gff_record_into_database(self, gff_record: gffutils.Feature, organism_entry: organism.Organism,
                                         all_feature_entries: Dict[str, sequence.Feature]):
        """Inserts, updates or deletes entries in various tables"""

        # Insert/update/get entry in the 'feature' tables
        feature_entry = self._handle_child_feature(gff_record, organism_entry)
        if feature_entry:

            # Insert/update/delete entries connected to this 'feature' entry in various tables
            self._handle_location(gff_record, feature_entry)
            self._handle_synonyms(gff_record, feature_entry)
            self._handle_properties(gff_record, feature_entry)
            self._handle_cross_references(gff_record, feature_entry)
            self._handle_ontology_terms(gff_record, feature_entry)
            self._handle_publications(gff_record, feature_entry)
            self._handle_relationships(gff_record, feature_entry, all_feature_entries)

            # Insert/update/delete entries connected to the associated protein (if present) in various tables
            self._handle_protein(gff_record, feature_entry, organism_entry, all_feature_entries)

            # Save 'feature' entry in global array
            self._check_if_gff_attributes_are_recognized(gff_record)
            all_feature_entries[feature_entry.uniquename] = feature_entry

    def _handle_child_feature(self, gff_record: gffutils.Feature, organism_entry: organism.Organism
                              ) -> Union[None, sequence.Feature]:
        """Inserts or updates an entry in the 'feature' table and returns it"""

        # Check if all dependencies are met
        if gff_record.featuretype not in self._sequence_terms:
            self.printer.print("WARNING: Sequence type '" + gff_record.featuretype + "' not present in database")
            return None
        type_entry = self._sequence_terms[gff_record.featuretype]

        # Create a feature object, and update the corresponding table
        new_feature_entry = self._create_feature(gff_record, organism_entry.organism_id, type_entry.cvterm_id)
        feature_entry = self._handle_feature(new_feature_entry, organism_entry.abbreviation)
        return feature_entry

    def _handle_location(self, gff_record: gffutils.Feature, feature_entry: sequence.Feature
                         ) -> Union[None, sequence.FeatureLoc]:
        """Inserts or updates an entry in the 'featureloc' table and returns it"""

        # Get entry from 'srcfeature' table
        srcfeature_entry = self.query_first(sequence.Feature, organism_id=feature_entry.organism_id,
                                            uniquename=gff_record.seqid)
        if not srcfeature_entry:
            self.printer.print("WARNING: Parent sequence '" + gff_record.seqid + "' not present in database")
            return None

        # Insert/update entry in the 'featureloc' table
        new_featureloc_entry = self._create_featureloc(gff_record, feature_entry.feature_id,
                                                       srcfeature_entry.feature_id)
        featureloc_entry = self._handle_featureloc(new_featureloc_entry, feature_entry.uniquename)
        return featureloc_entry

    def _handle_synonyms(self, gff_record: gffutils.Feature, feature_entry: sequence.Feature
                         ) -> List[sequence.FeatureSynonym]:
        """Inserts, updates and deletes entries in the 'synonym' and 'feature_synonym' tables and returns the latter"""

        # Extract existing synonyms for this feature from the database
        existing_feature_synonyms = self.query_feature_synonym_by_type(
            feature_entry.feature_id, self._synonym_type_ids).all()
        all_feature_synonyms = []

        # Loop over all synonyms for this feature in the GFF record
        synonyms = self._extract_gff_synonyms(gff_record)
        for synonym_type, aliases in synonyms.items():

            # Get database entry for synonym type
            if synonym_type not in self._synonym_terms:
                self.printer.print("WARNING: Synonym type '" + synonym_type + "' not present in database.")
                continue
            type_entry = self._synonym_terms[synonym_type]

            for alias in aliases:

                # Insert/update entry in the 'synonym' table
                new_synonym_entry = sequence.Synonym(name=alias, type_id=type_entry.cvterm_id, synonym_sgml=alias)
                synonym_entry = self._handle_synonym(new_synonym_entry)

                # Insert/update entry in the 'feature_synonym' table
                new_feature_synonym_entry = sequence.FeatureSynonym(
                    synonym_id=synonym_entry.synonym_id, feature_id=feature_entry.feature_id,
                    pub_id=self._default_pub.pub_id)
                feature_synonym_entry = self._handle_feature_synonym(new_feature_synonym_entry,
                                                                     existing_feature_synonyms)
                all_feature_synonyms.append(feature_synonym_entry)

        # Delete obsolete entries
        self._delete_feature_synonym(all_feature_synonyms, existing_feature_synonyms, feature_entry.uniquename)
        return all_feature_synonyms

    def _handle_publications(self, gff_record: gffutils.Feature, feature_entry: sequence.Feature
                             ) -> List[sequence.FeaturePub]:
        """Inserts, updates and deletes entries in the 'pub' and 'feature_pub' tables and returns the latter"""

        # Extract existing publications for this feature from the database
        existing_feature_pubs = self.query_all(sequence.FeaturePub, feature_id=feature_entry.feature_id)
        all_feature_pubs = []

        # Loop over all publications for this feature in the GFF record
        publications = self._extract_gff_publications(gff_record)
        for publication in publications:

            # Insert/update entry in the 'pub' table
            new_pub_entry = pub.Pub(uniquename=publication, type_id=self._default_pub.type_id)
            pub_entry = self._handle_pub(new_pub_entry)

            # Insert/update entry in the 'feature_pub' table
            new_feature_pub_entry = sequence.FeaturePub(feature_id=feature_entry.feature_id, pub_id=pub_entry.pub_id)
            feature_pub_entry = self._handle_feature_pub(new_feature_pub_entry, existing_feature_pubs,
                                                         feature_entry.uniquename, pub_entry.uniquename)
            all_feature_pubs.append(feature_pub_entry)

        # Delete obsolete entries
        self._delete_feature_pub(all_feature_pubs, existing_feature_pubs, feature_entry.uniquename)
        return all_feature_pubs

    def _handle_relationships(self, gff_record: gffutils.Feature, subject_entry: sequence.Feature,
                              all_features: Dict[str, sequence.Feature]) -> List[sequence.FeatureRelationship]:
        """Inserts, updates and deletes entries in the 'feature_relationship' table and returns them"""

        # Extract existing relationships for this subject from the database
        existing_feature_relationships = self.query_feature_relationship_by_type(
            subject_entry.feature_id, self._parent_type_ids).all()
        all_feature_relationships = []

        # Loop over all relationships for this feature in the GFF record
        relationships = self._extract_gff_relationships(gff_record)
        for relationship, parents in relationships.items():

            # Get database entry for relationship type
            type_entry = self._parent_terms[relationship]

            for parent in parents:

                # Get database entry for object
                if parent in all_features:
                    object_entry = all_features[parent]
                else:
                    object_entry = self.query_first(sequence.Feature, organism_id=subject_entry.organism_id,
                                                    uniquename=parent)
                if not object_entry:
                    self.printer.print("WARNING: Feature '" + parent +
                                       "' neither present in input file nor in database.")
                    continue

                # Insert/update entry in the 'feature_relationship' table
                new_relationship_entry = sequence.FeatureRelationship(subject_id=subject_entry.feature_id,
                                                                      object_id=object_entry.feature_id,
                                                                      type_id=type_entry.cvterm_id)
                feature_relationship_entry = self._handle_feature_relationship(
                    new_relationship_entry, existing_feature_relationships, subject_entry.uniquename,
                    object_entry.uniquename, type_entry.name)
                all_feature_relationships.append(feature_relationship_entry)

        # Delete obsolete entries
        self._delete_feature_relationship(all_feature_relationships, existing_feature_relationships,
                                          subject_entry.uniquename)
        return all_feature_relationships

    def _handle_properties(self, gff_record: gffutils.Feature, feature_entry: sequence.Feature
                           ) -> List[sequence.FeatureProp]:
        """Inserts, updates and deletes entries in the 'featureprop' table and returns them"""

        # Extract existing properties for this feature from the database
        existing_featureprops = self.query_featureprop_by_type(
            feature_entry.feature_id, self._feature_property_type_ids).all()
        all_featureprops = []

        # Loop over all properties of this feature in the GFF record
        props = self._extract_gff_properties(gff_record)
        for prop, values in props.items():
            for value in values:

                # Get database entry for property type from array
                if prop not in self._feature_property_terms:
                    self.printer.print("WARNING: Feature property term '" + prop + "' not present in input file.")
                    continue
                type_entry = self._feature_property_terms[prop]

                # Insert/update entry in the 'feature_relationship' table
                new_featureprop_entry = sequence.FeatureProp(feature_id=feature_entry.feature_id,
                                                             type_id=type_entry.cvterm_id, value=value)
                featureprop_entry = self._handle_featureprop(new_featureprop_entry, existing_featureprops, prop, value,
                                                             feature_entry.uniquename)
                all_featureprops.append(featureprop_entry)

        # Delete obsolete entries
        self._delete_featureprop(all_featureprops, existing_featureprops, feature_entry.uniquename)
        return all_featureprops

    def _handle_cross_references(self, gff_record: gffutils.Feature, feature_entry: sequence.Feature
                                 ) -> List[sequence.FeatureDbxRef]:
        """Inserts, updates and deletes entries in the 'feature_dbxref' table and returns them"""

        # Extract existing cross references for this feature from the database
        existing_feature_dbxrefs = self.query_all(sequence.FeatureDbxRef, feature_id=feature_entry.feature_id)
        all_feature_dbxrefs = []

        # Loop over all cross references of this feature in the GFF record
        crossrefs = self._extract_gff_crossrefs(gff_record)
        for crossref in crossrefs:

            # Split database cross reference (dbxref) into db, accession, version
            (db_authority, accession, version) = ontology.split_dbxref(crossref)

            # Insert/update entry in the 'db' table
            new_db_entry = general.Db(name=db_authority)
            db_entry = self._handle_db(new_db_entry)

            # Insert/update entry in the 'dbxref' table
            new_dbxref_entry = general.DbxRef(db_id=db_entry.db_id, accession=accession, version=version)
            dbxref_entry = self._handle_dbxref(new_dbxref_entry, db_authority)

            # Insert/update entry in the 'feature_dbxref' table
            new_feature_dbxref_entry = sequence.FeatureDbxRef(feature_id=feature_entry.feature_id,
                                                              dbxref_id=dbxref_entry.dbxref_id)
            feature_dbxref_entry = self._handle_feature_dbxref(new_feature_dbxref_entry, existing_feature_dbxrefs,
                                                               crossref, feature_entry.uniquename)
            all_feature_dbxrefs.append(feature_dbxref_entry)

        # Delete obsolete entries
        self._delete_feature_dbxref(all_feature_dbxrefs, existing_feature_dbxrefs, feature_entry.uniquename)
        return all_feature_dbxrefs

    def _handle_ontology_terms(self, gff_record: gffutils.Feature, feature_entry: sequence.Feature
                               ) -> List[sequence.FeatureCvTerm]:
        """Inserts, updates and deletes entries in the 'feature_cvterm' table and returns them"""

        # Extract existing ontology terms for this feature from the database
        existing_feature_cvterms = self.query_feature_cvterm_by_ontology(
            feature_entry.feature_id, self._go_db.db_id).all()
        all_feature_cvterms = []

        # Loop over all ontology terms of this feature in the GFF record
        ontology_terms = self._extract_gff_ontology_terms(gff_record)
        for ontology_term in ontology_terms:

            # Split database cross reference (dbxref) into db, accession, version
            (db_authority, accession, version) = ontology.split_dbxref(ontology_term)

            # Get entry from 'db' table
            db_entry = self.query_first(general.Db, name=db_authority)
            if not db_entry:
                self.printer.print("WARNING: Ontology '" + db_authority + "' not present in database.")
                continue

            # Get entry from 'dbxref' table
            dbxref_entry = self.query_first(general.DbxRef, db_id=db_entry.db_id, accession=accession)
            if not dbxref_entry:
                self.printer.print("WARNING: Ontology term '" + ontology_term + "' not present in database.")
                continue

            # Get entry from 'cvterm' table
            cvterm_entry = self.query_first(cv.CvTerm, dbxref_id=dbxref_entry.dbxref_id)
            if not cvterm_entry:
                self.printer.print("WARNING: CV term for ontology term '" + ontology_term
                                   + "' not present in database.")
                continue

            # Insert/update entry in the 'feature_cvterm' table
            new_feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=feature_entry.feature_id,
                                                              cvterm_id=cvterm_entry.cvterm_id,
                                                              pub_id=self._default_pub.pub_id)
            feature_cvterm_entry = self._handle_feature_cvterm(new_feature_cvterm_entry, existing_feature_cvterms,
                                                               cvterm_entry.name, feature_entry.uniquename)
            all_feature_cvterms.append(feature_cvterm_entry)

        # Delete obsolete entries
        self._delete_feature_cvterm(all_feature_cvterms, existing_feature_cvterms, feature_entry.uniquename)
        return all_feature_cvterms

    def _handle_protein(self, gff_record: gffutils.Feature, feature_entry: sequence.Feature,
                        organism_entry: organism.Organism, all_feature_entries: Dict[str, sequence.Feature]):
        """Creates a separate feature in Chado for a protein associated with a GFF feature"""

        # Check if the GFF record is associated with a polypeptide, and if yes extract its name
        protein_source_id = self._extract_protein_source_id(gff_record)
        if not protein_source_id:
            return

        # Get the 'feature' and 'featureloc' entries for the transcript
        if gff_record.featuretype.lower() in self._transcript_types():
            parent_entry = feature_entry
        else:
            parent_entry = self.query_parent_features(
                feature_entry.feature_id,
                [self._parent_terms["part_of"].cvterm_id]).first()                          # type: sequence.Feature
        if parent_entry:
            loc_entry = self.query_first(sequence.FeatureLoc,
                                         feature_id=parent_entry.feature_id)                # type: sequence.FeatureLoc
        else:
            loc_entry = sequence.FeatureLoc(feature_id=0, srcfeature_id=0, fmin=None, fmax=None, strand=None)

        # Create a new GFF record
        protein_feature = gffutils.Feature(seqid=gff_record.seqid, source=gff_record.source,
                                           start=loc_entry.fmin + 1, end=loc_entry.fmax,
                                           strand=self.back_convert_strand(loc_entry.strand),
                                           featuretype="polypeptide", id=protein_source_id,
                                           attributes={"Derives_from": parent_entry.uniquename})

        # Insert, update or delete entries in various tables for this GFF record (Note: recursive call)
        self._insert_gff_record_into_database(protein_feature, organism_entry, all_feature_entries)

    def _check_if_gff_attributes_are_recognized(self, gff_record: gffutils.Feature) -> bool:
        """Checks if all attributes of a GFF record can be recognized"""
        all_recognized = True
        for key in gff_record.attributes.keys():
            if key.lower() not in self._recognized_gff_attributes():
                self.printer.print("WARNING: The attribute '" + key + "' for feature '" + gff_record.id +
                                   "' is unknown by this program and can't be loaded into the database.")
                all_recognized = False
        return all_recognized

    def _create_feature(self, gff_record: gffutils.Feature, organism_id: int, type_id: int) -> sequence.Feature:
        """Creates a feature object from a GFF record"""
        name = self._extract_gff_name(gff_record)
        residues = self._extract_gff_translation(gff_record)
        seqlen = self._extract_gff_size(gff_record)
        if residues is not None:
            seqlen = len(residues)
        return sequence.Feature(organism_id=organism_id, type_id=type_id, uniquename=gff_record.id, name=name,
                                residues=residues, seqlen=seqlen)

    def _create_featureloc(self, gff_record: gffutils.Feature, feature_id: int, srcfeature_id: int
                           ) -> sequence.FeatureLoc:
        """Creates a featureloc object from a GFF record
        coordinates are converted from base-oriented (1-based) to interbase (0-based)"""
        return sequence.FeatureLoc(
            feature_id=feature_id, srcfeature_id=srcfeature_id, fmin=gff_record.start - 1, fmax=gff_record.end,
            strand=self.convert_strand(gff_record.strand), phase=self.convert_frame(gff_record.frame))

    @staticmethod
    def _extract_gff_crossrefs(gff_record: gffutils.Feature) -> List[str]:
        """Extract database cross references from a GFF record"""
        crossrefs = []
        for key, value in gff_record.attributes.items():
            if key.lower() == "dbxref":
                if isinstance(value, str):
                    crossrefs.append(value)
                elif isinstance(value, list):
                    crossrefs.extend(value)
        return crossrefs

    @staticmethod
    def _extract_gff_ontology_terms(gff_record: gffutils.Feature) -> List[str]:
        """Extract cross references to ontology terms from a GFF record"""
        ontology_terms = []
        if "Ontology_term" in gff_record.attributes:
            ontology_terms = gff_record.attributes["Ontology_term"]
            if isinstance(ontology_terms, str):
                ontology_terms = [ontology_terms]
        return ontology_terms

    def _extract_gff_properties(self, gff_record: gffutils.Feature) -> Dict[str, List[str]]:
        """Extracts properties from a GFF record. Properties are 'score', 'source' and certain attributes"""
        gff_to_chado_key = {"note": "comment"}
        properties = {}
        if gff_record.score and gff_record.score != '.':
            properties["score"] = [gff_record.score]
        if gff_record.source and gff_record.source != '.':
            properties["source"] = [gff_record.source]

        for gff_key, value in gff_record.attributes.items():

            chado_key = None
            if gff_key.lower() in gff_to_chado_key.keys():
                chado_key = gff_to_chado_key[gff_key.lower()]
            elif gff_key.lower() in self._feature_property_types():
                chado_key = gff_key.lower()

            if chado_key:
                if isinstance(value, str):
                    properties[chado_key] = [value]
                elif isinstance(value, list):
                    properties[chado_key] = value
        return properties

    def _extract_gff_relationships(self, gff_record: gffutils.Feature) -> Dict[str, List[str]]:
        """Extracts the relationships to other features from a GFF record"""
        gff_to_chado_key = {"parent": "part_of"}
        relationships = {}
        for gff_key, value in gff_record.attributes.items():

            chado_key = None
            if gff_key.lower() in gff_to_chado_key.keys():
                chado_key = gff_to_chado_key[gff_key.lower()]
            elif gff_key.lower() in self._feature_relationship_types():
                chado_key = gff_key.lower()

            if chado_key:
                if isinstance(value, str):
                    relationships[chado_key] = [value]
                elif isinstance(value, list):
                    relationships[chado_key] = value
        return relationships

    def _extract_gff_synonyms(self, gff_record: gffutils.Feature) -> Dict[str, List[str]]:
        """Extracts synonyms/alias names from a GFF record"""
        aliases = {}
        for gff_key, value in gff_record.attributes.items():

            chado_key = None
            if gff_key.lower() in self._synonym_types():
                chado_key = gff_key.lower()

            if chado_key:
                if isinstance(value, str):
                    aliases[gff_key.lower()] = [value]
                elif isinstance(value, list):
                    aliases[gff_key.lower()] = value
        return aliases

    @staticmethod
    def _extract_gff_name(gff_record: gffutils.Feature) -> Union[None, str]:
        """Extracts the feature name from a GFF record"""
        name = None
        if "Name" in gff_record.attributes:
            name = gff_record.attributes["Name"]
            if isinstance(name, list):
                name = name[0]
        return name

    @staticmethod
    def _extract_gff_size(gff_record: gffutils.Feature) -> Union[None, int]:
        """Extracts the sequence length of a feature from a GFF record"""
        size = None
        if "size" in gff_record.attributes:
            size = gff_record.attributes["size"]
            if isinstance(size, list):
                size = size[0]
            size = int(size)
        return size

    @staticmethod
    def _extract_gff_translation(gff_record: gffutils.Feature) -> Union[None, str]:
        """Extracts the sequence of amino acids from a GFF record representing a polypeptide"""
        residues = None
        if "translation" in gff_record.attributes:
            residues = gff_record.attributes["translation"]
            if isinstance(residues, str):
                residues = residues.upper()
            elif isinstance(residues, list):
                residues = residues[0].upper()
        return residues

    @staticmethod
    def _extract_gff_publications(gff_record: gffutils.Feature) -> List[str]:
        """Extracts publications names from a GFF record"""
        publications = []
        if "literature" in gff_record.attributes:
            publications = gff_record.attributes["literature"]
            if isinstance(publications, str):
                publications = [publications]
        return publications

    @staticmethod
    def _extract_protein_source_id(gff_record: gffutils.Feature) -> Union[None, str]:
        """Extracts the ID of a polypeptide associated with the feature from a GFF record"""
        protein_id = None
        if "protein_source_id" in gff_record.attributes:
            protein_id = gff_record.attributes["protein_source_id"]
            if isinstance(protein_id, list):
                protein_id = protein_id[0]
        return protein_id

    @staticmethod
    def _extract_gff_sequence_names(gff_db: gffutils.FeatureDB) -> List[str]:
        """Extracts sequence names from a GFF file"""
        sequences = []
        for directive in gff_db.directives:
            if directive.startswith("sequence-region"):
                split_directive = directive.split()
                sequence_name = split_directive[1].strip()
                sequences.append(sequence_name)
        return sequences

    def _recognized_gff_attributes(self) -> List[str]:
        """Lists GFF attributes that are handled by this program"""
        return self._feature_property_types() + self._feature_relationship_types() + self._synonym_types() \
            + ["id", "name", "parent", "note", "dbxref", "ontology_term", "size", "literature", "translation",
               "protein_source_id"]


class GFFExportClient(iobase.ChadoClient, GFFClient):
    """Class for exporting genomic data from Chado to GFF files"""

    def __init__(self, uri: str, verbose=False, test_environment=False):
        """Constructor"""

        # Connect to database
        self.uri = uri
        self.verbose = verbose
        self.test_environment = test_environment
        if self.test_environment:
            self.printer = utils.VerbosePrinter(self.verbose)
        else:
            super().__init__(self.uri, self.verbose)

        # Load essential database entries
        if not self.test_environment:
            self._load_essentials()

    def __del__(self):
        """Destructor - disconnect from database"""
        if not self.test_environment:
            super().__del__()

    def _load_essentials(self) -> None:
        """Loads essential database entries"""
        self._part_of_term = self._load_cvterm_from_cv("part_of", "relationship")
        self._derives_from_term = self._load_cvterm_from_cv("derives_from", "sequence")
        self._parent_terms = {self._part_of_term.name: self._part_of_term,
                              self._derives_from_term.name: self._derives_from_term}
        self._parent_type_ids = [self._part_of_term.cvterm_id, self._derives_from_term.cvterm_id]
        self._top_level_term = self._load_cvterm("top_level_seq")
        self._go_db = self._load_db("GO")

    def export(self, gff_filename: str, organism_name: str, export_fasta: bool, fasta_filename: str,
               include_obsolete_features=False) -> None:
        """Exports sequences from Chado to a GFF file"""

        # Load dependencies
        organism_entry = self._load_organism(organism_name)

        # Open GFF file
        gff_handle = utils.open_file_write(gff_filename)

        # Get top-level sequences and write a GFF header
        chromosome_entries = self.query_features_by_property_type(
            organism_entry.organism_id, self._top_level_term.cvterm_id).all()           # type: List[sequence.Feature]
        self._write_gff_header(gff_handle, chromosome_entries)

        # Loop over all top-level sequences
        for chromosome_entry in chromosome_entries:

            # Load all attributes associated with this sequence
            self._export_gff_record(chromosome_entry, chromosome_entry.uniquename, {}, gff_handle)

            # Get features located on this sequence
            feature_entries = self.query_features_by_srcfeature(chromosome_entry.feature_id).all()
            for feature_entry in feature_entries:

                # Create a GFF record for this feature, if it fulfills certain requirements
                if not self._has_feature_parents(feature_entry) \
                        and (include_obsolete_features or not feature_entry.is_obsolete):
                    self._export_gff_record(feature_entry, chromosome_entry.uniquename, {}, gff_handle)

        # Close GFF file
        utils.close(gff_handle)

        # Print FASTA sequences, if required
        if export_fasta:
            self._export_fasta(gff_filename, fasta_filename, organism_name)

    def _export_fasta(self, gff_file: str, fasta_file: str, organism_name: str) -> None:
        """Exports sequences from the Chado database into a FASTA file"""
        fasta_is_temporary = (fasta_file == "" or fasta_file is None)
        if fasta_is_temporary:

            # Create temporary file
            (fasta_handle, fasta_file) = tempfile.mkstemp(dir=os.getcwd())
            os.close(fasta_handle)

        # Export FASTA sequences to file
        fasta_client = fasta.FastaExportClient(self.uri, self.verbose)
        fasta_client.export(fasta_file, organism_name, "contigs", "")
        if fasta_is_temporary:

            # Append sequences to GFF and remove temporary file
            self._append_fasta(gff_file, fasta_file)
            os.remove(fasta_file)

    @staticmethod
    def _append_fasta(gff_file: str, fasta_file: str) -> None:
        """Appends FASTA sequences to a GFF file"""
        with open(gff_file, 'a') as gff_handle:
            gff_handle.write("##FASTA\n")
            with open(fasta_file, 'r') as fasta_handle:
                for line in fasta_handle:
                    gff_handle.write(line)

    @staticmethod
    def _write_gff_header(file_handle, chromosome_entries: List[sequence.Feature]) -> None:
        """Prints the header of a GFF file"""
        file_handle.write("##gff-version 3\n")
        file_handle.write("# feature-ontology so.obo\n")
        for chromosome_entry in chromosome_entries:
            seq = "\t".join(["##sequence-region", chromosome_entry.uniquename, "1",
                            str(chromosome_entry.seqlen)]) + "\n"
            file_handle.write(seq)

    @staticmethod
    def _print_gff_record(gff_record: gffutils.Feature, file_handle) -> None:
        """Prints a GFF record to file"""
        file_handle.write(str(gff_record) + "\n")

    def _export_gff_record(self, feature_entry: sequence.Feature, chromosome_name: str,
                           parent_relationships: Dict[str, str], file_handle) -> None:
        """Exports a GFF record for a given feature"""

        # Create GFF record
        gff_record = self._create_gff_record(feature_entry, chromosome_name)

        # Query various tables to gather information related to the feature
        feature_type = self._extract_feature_type(feature_entry)
        feature_synonyms = self._extract_feature_synonyms(feature_entry)
        feature_properties = self._extract_feature_properties(feature_entry)
        feature_publications = self._extract_feature_publications(feature_entry)
        feature_dbxrefs = self._extract_feature_cross_references(feature_entry)
        feature_ontology_terms = self._extract_feature_ontology_terms(feature_entry)

        # Add attributes to the GFF record
        self._add_gff_relationships(gff_record, parent_relationships)
        self._add_gff_cross_references(gff_record, feature_dbxrefs)
        self._add_gff_ontology_terms(gff_record, feature_ontology_terms)
        self._add_gff_synonyms(gff_record, feature_synonyms)
        self._add_gff_publications(gff_record, feature_publications)
        self._add_gff_properties(gff_record, feature_properties)
        self._add_gff_featuretype(gff_record, feature_type, feature_entry.residues)

        # Write the generated GFF record to file
        self._print_gff_record(gff_record, file_handle)

        # Handle children of this feature (recursive call)
        self._handle_child_features(feature_entry, chromosome_name, file_handle)

    def _handle_child_features(self, feature_entry: sequence.Feature, chromosome_name: str, file_handle) -> None:
        """Export GFF records for all child features of a given feature"""
        for relationship_type, relationship_term in self._parent_terms.items():
            child_entries = self.query_child_features(feature_entry.feature_id, relationship_term.cvterm_id).all()
            parent_relationships = {relationship_type: feature_entry.uniquename}
            for child_entry in child_entries:
                if not child_entry.is_obsolete:
                    self._export_gff_record(child_entry, chromosome_name, parent_relationships, file_handle)

    def _has_feature_parents(self, feature_entry: sequence.Feature) -> bool:
        """Checks if a given Feature has parents in the database"""
        parent_entry = self.query_parent_features(feature_entry.feature_id, self._parent_type_ids).first()
        return parent_entry is not None

    def _extract_feature_type(self, feature_entry: sequence.Feature) -> str:
        """Extracts the type of a feature by a database query"""
        cvterm_entry = self.query_first(cv.CvTerm, cvterm_id=feature_entry.type_id)
        return cvterm_entry.name

    def _extract_feature_synonyms(self, feature_entry: sequence.Feature) -> Dict[str, List[str]]:
        """Extracts synonyms of a feature by a database query"""
        synonyms = {}
        for synonym_type, synonym_name in self.query_feature_synonyms(feature_entry.feature_id).all():
            if synonym_type in synonyms:
                synonyms[synonym_type].append(synonym_name)
            else:
                synonyms[synonym_type] = [synonym_name]
        return synonyms

    def _extract_feature_properties(self, feature_entry: sequence.Feature) -> Dict[str, List[str]]:
        """Extracts properties of a feature by a database query"""
        properties = {}
        for property_type, property_value in self.query_feature_properties(feature_entry.feature_id).all():
            if property_type in properties:
                properties[property_type].append(property_value)
            else:
                properties[property_type] = [property_value]
        return properties

    def _extract_feature_publications(self, feature_entry: sequence.Feature) -> List[str]:
        """Extracts publications associated with a feature by a database query"""
        publications = []
        for publication, in self.query_feature_pubs(feature_entry.feature_id).all():
            publications.append(publication)
        return publications

    def _extract_feature_cross_references(self, feature_entry: sequence.Feature) -> List[str]:
        """Extracts cross references associated with a feature by a database query"""
        cross_references = []
        for db_authority, accession in self.query_feature_dbxrefs(feature_entry.feature_id).all():
            crossref = ontology.create_dbxref(db_authority, accession)
            cross_references.append(crossref)
        return cross_references

    def _extract_feature_ontology_terms(self, feature_entry: sequence.Feature) -> List[str]:
        """Extracts ontology terms associated with a feature by a database query"""
        ontology_terms = []
        for db_authority, accession in self.query_feature_ontology_terms(
                feature_entry.feature_id, self._go_db.db_id).all():
            crossref = ontology.create_dbxref(db_authority, accession)
            ontology_terms.append(crossref)
        return ontology_terms

    def _create_gff_record(self, feature_entry: sequence.Feature, chromosome_name: str) -> gffutils.Feature:
        """Creates a GFF record with the fields 'seqid', 'start', 'end', 'strand' and 'phase'"""
        featureloc_entry = self.query_first(sequence.FeatureLoc, feature_id=feature_entry.feature_id)
        if featureloc_entry:
            gff_record = gffutils.Feature(seqid=chromosome_name, id=feature_entry.uniquename,
                                          start=featureloc_entry.fmin+1, end=featureloc_entry.fmax,
                                          strand=self.back_convert_strand(featureloc_entry.strand),
                                          frame=self.back_convert_frame(featureloc_entry.phase))
        else:
            gff_record = gffutils.Feature(seqid=chromosome_name, id=feature_entry.uniquename,
                                          start=1, end=feature_entry.seqlen)
        gff_record.attributes["ID"] = feature_entry.uniquename
        if feature_entry.name:
            gff_record.attributes["Name"] = feature_entry.name
        return gff_record

    @staticmethod
    def _add_gff_featuretype(gff_record: gffutils.Feature, featuretype: str, residues: str) -> None:
        """Adds the 'type' and potentially the attribute 'translation' to a GFF record"""
        gff_record.featuretype = featuretype
        if gff_record.featuretype == "polypeptide" and residues:
            gff_record.attributes["translation"] = residues.upper()

    @staticmethod
    def _add_gff_synonyms(gff_record: gffutils.Feature, synonyms: Dict[str, List[str]]):
        """Adds the attribute 'Alias' and various other attributes to a GFF record"""
        for synonym_type in ["Alias", "synonym", "previous_systematic_id"]:
            if synonym_type.lower() in synonyms:
                gff_record.attributes[synonym_type] = synonyms[synonym_type.lower()]

    @staticmethod
    def _add_gff_properties(gff_record: gffutils.Feature, properties: Dict[str, List[str]]) -> None:
        """Adds the 'source', 'score' and various attributes to a GFF record"""
        if "score" in properties:
            gff_record.score = properties["score"][0]
        if "source" in properties:
            gff_record.source = properties["source"][0]
        for property_type in ["comment", "description", "Note"]:
            if property_type.lower() in properties:
                gff_record.attributes[property_type] = properties[property_type.lower()]

    @staticmethod
    def _add_gff_publications(gff_record: gffutils.Feature, publications: List[str]):
        """Adds the attribute 'literature' to a GFF record"""
        if publications:
            gff_record.attributes["literature"] = publications

    @staticmethod
    def _add_gff_cross_references(gff_record: gffutils.Feature, cross_references: List[str]):
        """Adds the attribute 'Dbxref' to a GFF record"""
        if cross_references:
            gff_record.attributes["Dbxref"] = cross_references

    @staticmethod
    def _add_gff_ontology_terms(gff_record: gffutils.Feature, ontology_terms: List[str]):
        """Adds the attribute 'Ontology_term' to a GFF record"""
        if ontology_terms:
            gff_record.attributes["Ontology_term"] = ontology_terms

    @staticmethod
    def _add_gff_relationships(gff_record: gffutils.Feature, relationships: Dict[str, str]) -> None:
        """Adds the attributes 'Parent' and 'Derives_from' to a GFF record"""
        chado_to_gff_key = {"part_of": "Parent", "derives_from": "Derives_from"}
        for chado_relationship_type, gff_relationship_type in chado_to_gff_key.items():
            if chado_relationship_type in relationships:
                gff_record.attributes[gff_relationship_type] = relationships[chado_relationship_type]



import unittest.mock
import os
import tempfile
import filecmp
from .. import utils
from ..io import iobase, gaf
from ..orm import general, cv, pub, organism, sequence

modules_dir = os.path.dirname(os.path.abspath(gaf.__file__))
data_dir = os.path.abspath(os.path.join(modules_dir, '..', 'tests', 'data'))


class TestGAFImport(unittest.TestCase):
    """Tests various functions used to load a GFF file into a database"""

    @classmethod
    def setUpClass(cls):
        # Creates an instance of the base class to be tested and instantiates global attributes
        cls.client = gaf.GAFImportClient("testuri", test_environment=True)
        cls.client.printer = utils.VerbosePrinter(False)
        cls.client._date_term = cv.CvTerm(cv_id=11, dbxref_id=91, name="date", cvterm_id=91)
        cls.client._evidence_term = cv.CvTerm(cv_id=11, dbxref_id=92, name="evidence", cvterm_id=92)
        cls.client._synonym_term = cv.CvTerm(cv_id=11, dbxref_id=93, name="synonym", cvterm_id=93)
        cls.client._default_pub = pub.Pub(uniquename="null", type_id=71, pub_id=33)
        cls.client._go_db = general.Db(name="GO", db_id=44)

    def setUp(self):
        # Creates a default GAF record
        self.default_gaf_record = {'DB': 'testdb', 'DB_Object_ID': 'testname', 'DB_Object_Symbol': 'testsymbol',
                                   'Qualifier': [''], 'GO_ID': 'GO:12345', 'DB:Reference': ['testdb:testaccession'],
                                   'Evidence': 'TAS', 'With': ['evidencedb:evidenceaccession'], 'Aspect': 'C',
                                   'DB_Object_Name': 'testproduct', 'Synonym': ['S1', 'S2'],
                                   'DB_Object_Type': 'transcript', 'Taxon_ID': ['testtaxon'],
                                   'Date': 'testdate', 'Assigned_By': 'testdb'}

    @unittest.mock.patch("pychado.io.gaf.GAFImportClient.query_first")
    def test_load_feature(self, mock_query: unittest.mock.Mock):
        # Tests the function loading the entry from the 'feature' table that corresponds to a GAF record
        self.assertIs(mock_query, self.client.query_first)
        organism_entry = organism.Organism(genus="", species="", abbreviation="testorganism", organism_id=1)
        self.client._load_feature(self.default_gaf_record, organism_entry)
        mock_query.assert_called_with(sequence.Feature, organism_id=1, uniquename="testname")

    def test_handle_name(self):
        # Tests the function that updates a feature name according to a GAF record
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        self.client._handle_name(self.default_gaf_record, feature_entry)
        self.assertEqual(feature_entry.name, "testsymbol")
        self.default_gaf_record["DB_Object_Symbol"] = "testname"
        self.client._handle_name(self.default_gaf_record, feature_entry)
        self.assertNotEqual(feature_entry.name, "testname")

    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_feature_synonym")
    @unittest.mock.patch("pychado.orm.sequence.FeatureSynonym")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_synonym")
    @unittest.mock.patch("pychado.orm.sequence.Synonym")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient.query_feature_synonym_by_type")
    def test_handle_synonyms(self, mock_query: unittest.mock.Mock, mock_synonym: unittest.mock.Mock,
                             mock_insert_synonym: unittest.mock.Mock, mock_feature_synonym: unittest.mock.Mock,
                             mock_insert_feature_synonym: unittest.mock.Mock):
        # Tests the function transferring data from a GAF record to the 'feature_synonym' table
        self.assertIs(mock_query, self.client.query_feature_synonym_by_type)
        self.assertIs(mock_synonym, sequence.Synonym)
        self.assertIs(mock_insert_synonym, self.client._handle_synonym)
        self.assertIs(mock_feature_synonym, sequence.FeatureSynonym)
        self.assertIs(mock_insert_feature_synonym, self.client._handle_feature_synonym)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        mock_query.return_value.configure_mock(**{"all.return_value": []})
        mock_insert_synonym.return_value = utils.EmptyObject(synonym_id=99)

        all_feature_synonyms = self.client._handle_synonyms(self.default_gaf_record, feature_entry)
        mock_query.assert_called_with(12, [93])
        mock_synonym.assert_any_call(name="S1", type_id=93, synonym_sgml="S1")
        mock_feature_synonym.assert_any_call(synonym_id=99, feature_id=12, pub_id=33)
        self.assertEqual(len(all_feature_synonyms), 2)

    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_feature_cvterm")
    @unittest.mock.patch("pychado.orm.sequence.FeatureCvTerm")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_pub")
    @unittest.mock.patch("pychado.orm.pub.Pub")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._extract_primary_publication")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient.query_first")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient.query_feature_cvterm_by_ontology")
    def test_handle_ontology_term(self, mock_query: unittest.mock.Mock, mock_query_first: unittest.mock.Mock,
                                  mock_extract: unittest.mock.Mock,
                                  mock_pub: unittest.mock.Mock, mock_insert_pub: unittest.mock.Mock,
                                  mock_feature_cvterm: unittest.mock.Mock,
                                  mock_insert_feature_cvterm: unittest.mock.Mock):
        # Tests the function transferring data from a GAF record to the 'feature_cvterm' table
        self.assertIs(mock_query, self.client.query_feature_cvterm_by_ontology)
        self.assertIs(mock_query_first, self.client.query_first)
        self.assertIs(mock_extract, self.client._extract_primary_publication)
        self.assertIs(mock_pub, pub.Pub)
        self.assertIs(mock_insert_pub, self.client._handle_pub)
        self.assertIs(mock_feature_cvterm, sequence.FeatureCvTerm)
        self.assertIs(mock_insert_feature_cvterm, self.client._handle_feature_cvterm)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        mock_extract.return_value = "PMID:12345"
        mock_query_first.side_effect = [utils.EmptyObject(db_id=33),
                                        utils.EmptyObject(dbxref_id=44),
                                        utils.EmptyObject(cvterm_id=55, name="")]
        mock_insert_pub.return_value = utils.EmptyObject(pub_id=66)

        ontology_term = self.client._handle_ontology_term(self.default_gaf_record, feature_entry)
        mock_query_first.assert_any_call(general.DbxRef, db_id=33, accession="12345")
        mock_query_first.assert_any_call(cv.CvTerm, dbxref_id=44)
        self.assertEqual(mock_query_first.call_count, 3)
        mock_pub.assert_any_call(uniquename="PMID:12345", type_id=71)
        self.assertEqual(mock_insert_pub.call_count, 1)
        mock_query.assert_called_with(12, 33)
        mock_feature_cvterm.assert_any_call(feature_id=12, cvterm_id=55, pub_id=66, is_not=False)
        self.assertEqual(mock_insert_feature_cvterm.call_count, 1)
        self.assertIsNotNone(ontology_term)

    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_feature_cvterm")
    @unittest.mock.patch("pychado.orm.sequence.FeatureCvTerm")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_pub")
    @unittest.mock.patch("pychado.orm.pub.Pub")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_cvterm")
    @unittest.mock.patch("pychado.orm.cv.CvTerm")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_cv")
    @unittest.mock.patch("pychado.orm.cv.Cv")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_dbxref")
    @unittest.mock.patch("pychado.orm.general.DbxRef")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_db")
    @unittest.mock.patch("pychado.orm.general.Db")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._extract_primary_publication")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient.query_feature_cvterm_by_ontology")
    def test_handle_product_term(self, mock_query: unittest.mock.Mock, mock_extract: unittest.mock.Mock,
                                 mock_db: unittest.mock.Mock, mock_insert_db: unittest.mock.Mock,
                                 mock_dbxref: unittest.mock.Mock, mock_insert_dbxref: unittest.mock.Mock,
                                 mock_cv: unittest.mock.Mock, mock_insert_cv: unittest.mock.Mock,
                                 mock_cvterm: unittest.mock.Mock, mock_insert_cvterm: unittest.mock.Mock,
                                 mock_pub: unittest.mock.Mock, mock_insert_pub: unittest.mock.Mock,
                                 mock_feature_cvterm: unittest.mock.Mock,
                                 mock_insert_feature_cvterm: unittest.mock.Mock):
        # Tests the function transferring data from a GAF record to the 'feature_cvterm' table
        self.assertIs(mock_query, self.client.query_feature_cvterm_by_ontology)
        self.assertIs(mock_extract, self.client._extract_primary_publication)
        self.assertIs(mock_db, general.Db)
        self.assertIs(mock_insert_db, self.client._handle_db)
        self.assertIs(mock_dbxref, general.DbxRef)
        self.assertIs(mock_insert_dbxref, self.client._handle_dbxref)
        self.assertIs(mock_cv, cv.Cv)
        self.assertIs(mock_insert_cv, self.client._handle_cv)
        self.assertIs(mock_cvterm, cv.CvTerm)
        self.assertIs(mock_insert_cvterm, self.client._handle_cvterm)
        self.assertIs(mock_pub, pub.Pub)
        self.assertIs(mock_insert_pub, self.client._handle_pub)
        self.assertIs(mock_feature_cvterm, sequence.FeatureCvTerm)
        self.assertIs(mock_insert_feature_cvterm, self.client._handle_feature_cvterm)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        mock_extract.return_value = "PMID:12345"
        mock_insert_db.return_value = utils.EmptyObject(db_id=22, name="")
        mock_insert_dbxref.return_value = utils.EmptyObject(dbxref_id=33)
        mock_insert_cv.return_value = utils.EmptyObject(cv_id=44, name="")
        mock_insert_cvterm.return_value = utils.EmptyObject(cvterm_id=55, name="")
        mock_insert_pub.return_value = utils.EmptyObject(pub_id=66)

        product_term = self.client._handle_product_term(self.default_gaf_record, feature_entry)
        mock_db.assert_called_with(name="PRODUCT")
        mock_dbxref.assert_called_with(db_id=22, accession="testproduct")
        mock_cv.assert_called_with(name="genedb_products")
        mock_cvterm.assert_called_with(cv_id=44, dbxref_id=33, name="testproduct")
        mock_pub.assert_called_with(uniquename="PMID:12345", type_id=71)
        mock_query.assert_called_with(12, 22)
        mock_feature_cvterm.assert_called_with(feature_id=12, cvterm_id=55, pub_id=66)

        mock_insert_db.assert_called()
        mock_insert_dbxref.assert_called()
        mock_insert_cv.assert_called()
        mock_insert_cvterm.assert_called()
        mock_insert_pub.assert_called()
        mock_insert_feature_cvterm.assert_called()
        self.assertIsNotNone(product_term)

    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_feature_cvtermprop")
    @unittest.mock.patch("pychado.orm.sequence.FeatureCvTermProp")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient.query_all")
    def test_handle_properties(self, mock_query: unittest.mock.Mock, mock_prop: unittest.mock.Mock,
                               mock_insert_prop: unittest.mock.Mock):
        # Tests the function transferring data from a GAF record to the 'feature_cvtermprop' table
        self.assertIs(mock_query, self.client.query_all)
        self.assertIs(mock_prop, sequence.FeatureCvTermProp)
        self.assertIs(mock_insert_prop, self.client._handle_feature_cvtermprop)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=1, cvterm_id=2, pub_id=3, feature_cvterm_id=4)

        all_properties = self.client._handle_properties(self.default_gaf_record, feature_cvterm_entry, feature_entry)
        mock_query.assert_called_with(sequence.FeatureCvTermProp, feature_cvterm_id=4)
        mock_prop.assert_any_call(feature_cvterm_id=4, type_id=91, value="testdate")
        mock_prop.assert_any_call(feature_cvterm_id=4, type_id=92, value="Traceable Author Statement")
        self.assertEqual(mock_insert_prop.call_count, 2)
        self.assertEqual(len(all_properties), 2)

    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_feature_cvterm_dbxref")
    @unittest.mock.patch("pychado.orm.sequence.FeatureCvTermDbxRef")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_dbxref")
    @unittest.mock.patch("pychado.orm.general.DbxRef")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_db")
    @unittest.mock.patch("pychado.orm.general.Db")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._extract_cross_references")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient.query_all")
    def test_handle_crossrefs(self, mock_query: unittest.mock.Mock, mock_extract: unittest.mock.Mock,
                              mock_db: unittest.mock.Mock, mock_insert_db: unittest.mock.Mock,
                              mock_dbxref: unittest.mock.Mock, mock_insert_dbxref: unittest.mock.Mock,
                              mock_feature_cvterm_dbxref: unittest.mock.Mock,
                              mock_insert_feature_cvterm_dbxref: unittest.mock.Mock):
        # Tests the function transferring data from a GAF record to the 'feature_cvterm_dbxref' table
        self.assertIs(mock_query, self.client.query_all)
        self.assertIs(mock_extract, self.client._extract_cross_references)
        self.assertIs(mock_db, general.Db)
        self.assertIs(mock_insert_db, self.client._handle_db)
        self.assertIs(mock_dbxref, general.DbxRef)
        self.assertIs(mock_insert_dbxref, self.client._handle_dbxref)
        self.assertIs(mock_feature_cvterm_dbxref, sequence.FeatureCvTermDbxRef)
        self.assertIs(mock_insert_feature_cvterm_dbxref, self.client._handle_feature_cvterm_dbxref)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=1, cvterm_id=2, pub_id=3, feature_cvterm_id=4)
        mock_extract.return_value = ["evidencedb:evidenceaccession"]
        mock_insert_db.return_value = utils.EmptyObject(db_id=44, name="")
        mock_insert_dbxref.return_value = utils.EmptyObject(dbxref_id=55, accession="", version="")

        all_crossrefs = self.client._handle_crossrefs(self.default_gaf_record, feature_cvterm_entry, feature_entry)
        mock_query.assert_called_with(sequence.FeatureCvTermDbxRef, feature_cvterm_id=4)
        mock_db.assert_any_call(name="evidencedb")
        mock_dbxref.assert_any_call(db_id=44, accession="evidenceaccession", version="")
        mock_feature_cvterm_dbxref.assert_any_call(feature_cvterm_id=4, dbxref_id=55)
        self.assertEqual(mock_insert_db.call_count, 1)
        self.assertEqual(mock_insert_dbxref.call_count, 1)
        self.assertEqual(mock_insert_feature_cvterm_dbxref.call_count, 1)
        self.assertEqual(len(all_crossrefs), 1)

    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_feature_cvterm_pub")
    @unittest.mock.patch("pychado.orm.sequence.FeatureCvTermPub")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._handle_pub")
    @unittest.mock.patch("pychado.orm.pub.Pub")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient._extract_secondary_publications")
    @unittest.mock.patch("pychado.io.gaf.GAFImportClient.query_all")
    def test_handle_publications(self, mock_query: unittest.mock.Mock, mock_extract: unittest.mock.Mock,
                                 mock_pub: unittest.mock.Mock, mock_insert_pub: unittest.mock.Mock,
                                 mock_feature_cvterm_pub: unittest.mock.Mock,
                                 mock_insert_feature_cvterm_pub: unittest.mock.Mock):
        # Tests the function transferring data from a GAF record to the 'feature_cvterm_pub' table
        self.assertIs(mock_query, self.client.query_all)
        self.assertIs(mock_extract, self.client._extract_secondary_publications)
        self.assertIs(mock_pub, pub.Pub)
        self.assertIs(mock_insert_pub, self.client._handle_pub)
        self.assertIs(mock_feature_cvterm_pub, sequence.FeatureCvTermPub)
        self.assertIs(mock_insert_feature_cvterm_pub, self.client._handle_feature_cvterm_pub)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=1, cvterm_id=2, pub_id=3, feature_cvterm_id=4)
        mock_insert_pub.return_value = utils.EmptyObject(pub_id=33, uniquename="")
        mock_extract.return_value = []

        all_publications = self.client._handle_publications(self.default_gaf_record, feature_cvterm_entry,
                                                            feature_entry)
        mock_query.assert_called_with(sequence.FeatureCvTermPub, feature_cvterm_id=4)
        mock_pub.assert_not_called()
        self.assertEqual(len(all_publications), 0)

        mock_extract.return_value = ["new_publication"]
        all_publications = self.client._handle_publications(self.default_gaf_record, feature_cvterm_entry,
                                                            feature_entry)
        mock_pub.assert_called_with(uniquename="new_publication", type_id=71)
        self.assertEqual(mock_insert_pub.call_count, 1)
        mock_feature_cvterm_pub.assert_called_with(feature_cvterm_id=4, pub_id=33)
        self.assertEqual(mock_insert_feature_cvterm_pub.call_count, 1)
        self.assertEqual(len(all_publications), 1)

    def test_extract_cross_references(self):
        # Tests the function that extracts database cross references from a GAF record
        gaf_record = {"DB:Reference": ["somedb:someaccession", "PMID:12345", "another_publication"],
                      "With": ["evidencedb:evidenceaccession"]}
        crossrefs = self.client._extract_cross_references(gaf_record)
        self.assertEqual(crossrefs, ["evidencedb:evidenceaccession", "somedb:someaccession"])

    def test_extract_primary_publication(self):
        # Tests the function that extracts the primary publication from a GAF record
        gaf_record = {"DB:Reference": ["somedb:someaccession", "PMID:12345", "another_publication", "GO_REF:0001"]}
        primary_publication = self.client._extract_primary_publication(gaf_record)
        self.assertEqual(primary_publication, "PMID:12345")
        gaf_record = {"DB:Reference": ["somedb:someaccession", "another_publication", "GO_REF:0001"]}
        primary_publication = self.client._extract_primary_publication(gaf_record)
        self.assertEqual(primary_publication, "another_publication")
        gaf_record = {"DB:Reference": ["somedb:someaccession", "GO_REF:0001"]}
        primary_publication = self.client._extract_primary_publication(gaf_record)
        self.assertIsNone(primary_publication)

    def test_extract_secondary_publications(self):
        # Tests the function that extracts secondary publications from a GAF record
        gaf_record = {"DB:Reference": ["somedb:someaccession", "PMID:12345", "another_publication", "GO_REF:0001"]}
        secondary_publications = self.client._extract_secondary_publications(gaf_record)
        self.assertEqual(secondary_publications, ["another_publication"])
        gaf_record = {"DB:Reference": ["somedb:someaccession", "another_publication", "GO_REF:0001"]}
        secondary_publications = self.client._extract_secondary_publications(gaf_record)
        self.assertEqual(secondary_publications, [])


class TestGAFExport(unittest.TestCase):
    """Tests various functions used to export a GFF file from a database"""

    @classmethod
    def setUpClass(cls):
        # Creates an instance of the base class to be tested and instantiates global attributes
        cls.client = gaf.GAFExportClient("testuri", test_environment=True)
        cls.client._taxon_term = cv.CvTerm(cv_id=22, dbxref_id=93, name="taxonId", cvterm_id=93)
        cls.client._go_db = general.Db(name="GO", db_id=44)
        cls.client._product_db = general.Db(name="PRODUCT", db_id=55)

    @unittest.mock.patch("pychado.io.gaf.GAFExportClient.query_first")
    def test_extract_taxon_id(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts a taxon ID by a database query
        self.assertIs(mock_query, self.client.query_first)
        organism_entry = organism.Organism(genus="", species="", abbreviation="testorganism", organism_id=1)

        mock_query.return_value = None
        taxon = self.client._extract_taxon_id(organism_entry)
        mock_query.assert_called_with(organism.OrganismProp, organism_id=1, type_id=93)
        self.assertEqual(taxon, "taxon:testorganism")

        mock_query.return_value = organism.OrganismProp(organism_id=1, type_id=93, value="12345")
        taxon = self.client._extract_taxon_id(organism_entry)
        self.assertEqual(taxon, "taxon:12345")

    def test_write_gaf(self):
        # Tests the functions writing header and records to a GAF file
        gaf_file = tempfile.mkstemp()[1]
        gaf_file_handle = utils.open_file_write(gaf_file)
        gaf_record = {"k1": "v1", "k2": ["v2.1", "v2.2"]}
        self.client._write_gaf_header(gaf_file_handle)
        self.client._print_gaf_record(gaf_file_handle, gaf_record)
        utils.close(gaf_file_handle)
        actual_gaf_file = os.path.join(data_dir, 'gaf_records.gaf')
        self.assertTrue(filecmp.cmp(gaf_file, actual_gaf_file))
        os.remove(gaf_file)

    def test_stringify_gaf_attribute(self):
        # Tests the function that converts a list into a pipe-separated string
        string = self.client._stringify_gaf_attribute("bbb")
        self.assertEqual(string, "bbb")
        string = self.client._stringify_gaf_attribute(["a", "bbb"])
        self.assertEqual(string, "a|bbb")

    @unittest.mock.patch("pychado.io.gaf.GAFExportClient.query_feature_cvterm_ontology_terms")
    def test_extract_feature_cvterm_ontology_term(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the ontology term associated with a feature_cvterm by a database query
        self.assertIs(mock_query, self.client.query_feature_cvterm_ontology_terms)
        feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=12, cvterm_id=33, pub_id=44, feature_cvterm_id=77)
        mock_query.return_value.configure_mock(**{"first.return_value": ("somedb", "someaccession")})
        ontology_term = self.client._extract_feature_cvterm_ontology_term(feature_cvterm_entry)
        mock_query.assert_called_with(77, 44)
        self.assertEqual(ontology_term, "somedb:someaccession")

    @unittest.mock.patch("pychado.io.gaf.GAFExportClient.query_feature_cvterm_properties")
    def test_extract_feature_cvterm_properties(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the properties associated with a feature_cvterm by a database query
        self.assertIs(mock_query, self.client.query_feature_cvterm_properties)
        feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=12, cvterm_id=33, pub_id=44, feature_cvterm_id=77)
        mock_query.return_value.configure_mock(**{"all.return_value": [("k1", "v1"), ("k2", "v2")]})
        properties = self.client._extract_feature_cvterm_properties(feature_cvterm_entry)
        mock_query.assert_called_with(77)
        self.assertEqual(properties, {"k1": "v1", "k2": "v2"})

    @unittest.mock.patch("pychado.io.gaf.GAFExportClient.query_feature_cvterm_dbxrefs")
    def test_extract_feature_cvterm_dbxrefs(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the cross references associated with a feature_cvterm by a database query
        self.assertIs(mock_query, self.client.query_feature_cvterm_dbxrefs)
        feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=12, cvterm_id=33, pub_id=44, feature_cvterm_id=77)
        mock_query.return_value.configure_mock(**{"all.return_value": [("somedb", "someacc"), ("otherdb", "otheracc")]})
        dbxrefs = self.client._extract_feature_cvterm_dbxrefs(feature_cvterm_entry)
        mock_query.assert_called_with(77)
        self.assertEqual(dbxrefs, ["somedb:someacc", "otherdb:otheracc"])

    @unittest.mock.patch("pychado.io.gaf.GAFExportClient.query_feature_cvterm_secondary_pubs")
    @unittest.mock.patch("pychado.io.gaf.GAFExportClient.query_feature_cvterm_pubs")
    def test_extract_feature_cvterm_publications(self, mock_query_primary: unittest.mock.Mock,
                                                 mock_query_secondary: unittest.mock.Mock):
        # Tests the function that extracts the publications associated with a feature_cvterm by a database query
        self.assertIs(mock_query_primary, self.client.query_feature_cvterm_pubs)
        self.assertIs(mock_query_secondary, self.client.query_feature_cvterm_secondary_pubs)
        feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=12, cvterm_id=33, pub_id=44, feature_cvterm_id=77)
        mock_query_primary.return_value.configure_mock(**{"all.return_value": [("pub0",)]})
        mock_query_secondary.return_value.configure_mock(**{"all.return_value": [("pub1",), ("pub2",)]})
        publications = self.client._extract_feature_cvterm_publications(feature_cvterm_entry)
        mock_query_primary.assert_called_with(77)
        mock_query_secondary.assert_called_with(77)
        self.assertEqual(publications, ["pub0", "pub1", "pub2"])

    @unittest.mock.patch("pychado.io.gaf.GAFExportClient.query_cvterm_namespace")
    def test_extract_go_namespace(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the namespace of a feature_cvterm by a database query
        self.assertIs(mock_query, self.client.query_cvterm_namespace)
        feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=12, cvterm_id=33, pub_id=44, feature_cvterm_id=77)
        mock_query.return_value.configure_mock(**{"scalar.return_value": "somecv"})
        namespace = self.client._extract_go_namespace(feature_cvterm_entry)
        mock_query.assert_called_with(33)
        self.assertEqual(namespace, "somecv")

    def test_extract_feature_name(self):
        # Tests the function that extracts the name of a feature
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        featurename = self.client._extract_feature_name(feature_entry)
        self.assertEqual(featurename, "testname")
        feature_entry.name = "user_friendly_name"
        featurename = self.client._extract_feature_name(feature_entry)
        self.assertEqual(featurename, "user_friendly_name")

    @unittest.mock.patch("pychado.io.gaf.GAFExportClient.query_feature_synonyms")
    def test_extract_feature_synonyms(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the synonyms of a feature by a database query
        self.assertIs(mock_query, self.client.query_feature_synonyms)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        mock_query.return_value.configure_mock(**{"all.return_value": [("synonym", "s1"), ("alias", "s2")]})
        synonyms = self.client._extract_feature_synonyms(feature_entry)
        mock_query.assert_called_with(12)
        self.assertEqual(synonyms, ["s1", "s2"])

    @unittest.mock.patch("pychado.io.gaf.GAFExportClient.query_first")
    @unittest.mock.patch("pychado.io.gaf.GAFExportClient.query_feature_cvterm_by_ontology")
    def test_extract_product_name(self, mock_query: unittest.mock.Mock, mock_query_first: unittest.mock.Mock):
        # Tests the function that extracts the name of a gene product by a database query
        self.assertIs(mock_query, self.client.query_feature_cvterm_by_ontology)
        self.assertIs(mock_query_first, self.client.query_first)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        mock_query.return_value.configure_mock(**{"first.return_value": sequence.FeatureCvTerm(
            feature_id=12, cvterm_id=44, pub_id=0)})
        mock_query_first.return_value = cv.CvTerm(cv_id=1, dbxref_id=2, name="someproduct")
        product_name = self.client._extract_product_name(feature_entry)
        mock_query.assert_called_with(12, 55)
        mock_query_first.assert_called_with(cv.CvTerm, cvterm_id=44)
        self.assertEqual(product_name, "someproduct")

    def test_create_gaf_record(self):
        # Tests the function that creates a GAF record
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=12, cvterm_id=33, pub_id=44, is_not=True)
        gaf_record = self.client._create_gaf_record(feature_cvterm_entry, feature_entry, "testdb", "testtaxon")
        self.assertEqual(len(gaf_record), 15)
        self.assertIn("Evidence", gaf_record)
        self.assertEqual(gaf_record["DB"], "testdb")
        self.assertEqual(gaf_record["Assigned_By"], "testdb")
        self.assertEqual(gaf_record["DB_Object_ID"], "testname")
        self.assertEqual(gaf_record["Taxon_ID"], "testtaxon")
        self.assertEqual(gaf_record["Qualifier"], "NOT")

    def test_add_gaf_go_id(self):
        # Tests the function that adds a GO term to a GAF record
        gaf_record = {"DB_Object_ID": "objid"}
        self.client._add_gaf_go_id(gaf_record, "GO:12345")
        self.assertEqual(gaf_record["GO_ID"], "GO:12345")
        with self.assertRaises(iobase.DatabaseError):
            self.client._add_gaf_go_id(gaf_record, "")

    def test_add_gaf_aspect(self):
        # Tests the function that adds a GO namespace (aspect) to a GAF record
        gaf_record = {"DB_Object_ID": "objid", "GO_ID": "GO:12345"}
        self.client._add_gaf_aspect(gaf_record, "molecular_function")
        self.assertEqual(gaf_record["Aspect"], "F")
        with self.assertRaises(iobase.DatabaseError):
            self.client._add_gaf_aspect(gaf_record, "inexistent_namespace")

    def test_add_gaf_annotation_date(self):
        # Tests the function that adds an annotation date to a GAF record
        gaf_record = {"DB_Object_ID": "objid", "GO_ID": "GO:12345"}
        self.client._add_gaf_annotation_date(gaf_record, {"date": "20180101"})
        self.assertEqual(gaf_record["Date"], "20180101")
        self.client._add_gaf_annotation_date(gaf_record, {})
        self.assertEqual(gaf_record["Date"], utils.current_date())

    def test_add_gaf_evidence_code(self):
        # Tests the function that adds a GO evidence code to a GAF record
        gaf_record = {"DB_Object_ID": "objid", "GO_ID": "GO:12345"}
        self.client._add_gaf_evidence_code(gaf_record, {"evidence": "Inferred from High Throughput Experiment"})
        self.assertEqual(gaf_record["Evidence"], "HTP")
        self.client._add_gaf_evidence_code(gaf_record, {"evidence": "inexistent_evidence_code"})
        self.assertEqual(gaf_record["Evidence"], "NR")
        self.client._add_gaf_evidence_code(gaf_record, {})
        self.assertEqual(gaf_record["Evidence"], "NR")

    def test_add_gaf_withfrom_info(self):
        # Tests the functions that adds references for a GO evidence code to a GAF record
        gaf_record = {"DB_Object_ID": "objid"}
        self.client._add_gaf_withfrom_info(gaf_record, ["testdb:testaccession", "GO_REF:go_accession"])
        self.assertEqual(gaf_record["With"], ["testdb:testaccession"])

    def test_add_gaf_db_reference(self):
        # Tests the functions that adds database references to a GAF record
        gaf_record = {"DB_Object_ID": "objid"}
        self.client._add_gaf_db_references(gaf_record, ["PMID:12345", "null"],
                                           ["testdb:testaccession", "GO_REF:go_accession"])
        self.assertEqual(gaf_record["DB:Reference"], ["PMID:12345", "GO_REF:go_accession"])
        gaf_record.clear()
        self.client._add_gaf_db_references(gaf_record, [], [])
        self.assertEqual(gaf_record["DB:Reference"], ["GO_REF:0000002"])

    def test_add_gaf_object_type(self):
        # Tests the function that adds the feature type to a GAF record
        gaf_record = {"DB_Object_ID": "objid"}
        self.client._add_gaf_object_type(gaf_record, "testtype")
        self.assertEqual(gaf_record["DB_Object_Type"], "testtype")

    def test_add_gaf_object_name(self):
        # Tests the function that adds the name of the gene product to a GAF record
        gaf_record = {"DB_Object_ID": "objid"}
        self.client._add_gaf_object_name(gaf_record, "testproduct")
        self.assertEqual(gaf_record["DB_Object_Name"], "testproduct")

    def test_add_gaf_object_symbol(self):
        # Tests the function that adds the gene name to a GAF record
        gaf_record = {"DB_Object_ID": "objid"}
        self.client._add_gaf_object_symbol(gaf_record, "testname")
        self.assertEqual(gaf_record["DB_Object_Symbol"], "testname")
        with self.assertRaises(iobase.DatabaseError):
            self.client._add_gaf_object_symbol(gaf_record, "")

    def test_add_gaf_synonyms(self):
        # Tests the function that adds gene synonyms to a GAF record
        gaf_record = {"DB_Object_ID": "objid"}
        self.client._add_gaf_synonyms(gaf_record, ["s1", "s2"])
        self.assertEqual(gaf_record["Synonym"], ["s1", "s2"])


class TestGAFFunctions(unittest.TestCase):
    """Tests various general functions used in connection with GAF import/export"""

    @classmethod
    def setUpClass(cls):
        # Creates an instance of the base class to be tested
        cls.client = gaf.GAFClient("testuri", test_environment=True)
        cls.client._part_of_term = cv.CvTerm(cv_id=11, dbxref_id=91, name="part_of", is_relationshiptype=1,
                                             cvterm_id=91)
        cls.client._derives_from_term = cv.CvTerm(cv_id=11, dbxref_id=92, name="derives_from",
                                                  is_relationshiptype=1, cvterm_id=92)

    @unittest.mock.patch("pychado.io.gaf.GAFClient.query_first")
    def test_extract_feature_type(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the type of a feature by a database query
        self.assertIs(mock_query, self.client.query_first)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        mock_query.return_value = cv.CvTerm(cv_id=1, dbxref_id=2, name="sometype")
        featuretype = self.client._extract_feature_type(feature_entry)
        mock_query.assert_called_with(cv.CvTerm, cvterm_id=200)
        self.assertEqual(featuretype, "sometype")

    @unittest.mock.patch("pychado.io.gaf.GAFClient._extract_feature_type")
    def test_is_featuretype_valid(self, mock_extract: unittest.mock.Mock):
        # Tests the function that checks whether the type of a feature is valid
        self.assertIs(mock_extract, self.client._extract_feature_type)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        mock_extract.return_value = "mRNA"
        self.assertTrue(self.client._is_featuretype_valid(feature_entry))
        mock_extract.return_value = "chromosome"
        self.assertFalse(self.client._is_featuretype_valid(feature_entry))

    @unittest.mock.patch("pychado.io.gaf.GAFClient._extract_polypeptide_of_feature")
    @unittest.mock.patch("pychado.io.gaf.GAFClient._extract_transcript_of_feature")
    @unittest.mock.patch("pychado.io.gaf.GAFClient._extract_gene_of_feature")
    def test_extract_requested_feature(self, mock_gene: unittest.mock.Mock, mock_transcript: unittest.mock.Mock,
                                       mock_polypeptide: unittest.mock.Mock):
        # Tests the function that extracts a feature of a requested type from the database
        self.assertIs(mock_gene, self.client._extract_gene_of_feature)
        self.assertIs(mock_transcript, self.client._extract_transcript_of_feature)
        self.assertIs(mock_polypeptide, self.client._extract_polypeptide_of_feature)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)

        self.client._extract_requested_feature(feature_entry, "gene")
        mock_gene.assert_called_with(feature_entry)
        mock_transcript.assert_not_called()
        mock_polypeptide.assert_not_called()

        mock_gene.reset_mock()
        self.client._extract_requested_feature(feature_entry, "transcript")
        mock_gene.assert_not_called()
        mock_transcript.assert_called_with(feature_entry)
        mock_polypeptide.assert_not_called()

        mock_transcript.reset_mock()
        self.client._extract_requested_feature(feature_entry, "protein")
        mock_gene.assert_not_called()
        mock_transcript.assert_not_called()
        mock_polypeptide.assert_called_with(feature_entry)

        mock_polypeptide.reset_mock()
        requested_feature = self.client._extract_requested_feature(feature_entry, "default")
        mock_gene.assert_not_called()
        mock_transcript.assert_not_called()
        mock_polypeptide.assert_not_called()
        self.assertIs(requested_feature, feature_entry)

    @unittest.mock.patch("pychado.io.gaf.GAFClient.query_parent_features")
    @unittest.mock.patch("pychado.io.gaf.GAFClient._extract_feature_type")
    def test_extract_gene_of_feature(self, mock_type: unittest.mock.Mock, mock_parent: unittest.mock.Mock):
        # Tests the function that extracts the gene associated with a feature from the database
        self.assertIs(mock_type, self.client._extract_feature_type)
        self.assertIs(mock_parent, self.client.query_parent_features)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        other_feature_entry = sequence.Feature(organism_id=11, type_id=300, uniquename="othername", feature_id=13)

        # Feature is a gene - return input
        mock_type.return_value = "gene"
        gene_entry = self.client._extract_gene_of_feature(feature_entry)
        mock_type.assert_called_with(feature_entry)
        mock_parent.assert_not_called()
        self.assertIs(gene_entry, feature_entry)

        # Feature is a transcript - return parent
        mock_type.return_value = "tRNA"
        mock_parent.reset_mock()
        mock_parent.return_value.configure_mock(**{"first.return_value": other_feature_entry})
        gene_entry = self.client._extract_gene_of_feature(feature_entry)
        mock_parent.assert_called_with(12, [91])
        self.assertIs(gene_entry, other_feature_entry)

        # Feature is a polypeptide - return grandparent
        mock_type.return_value = "polypeptide"
        mock_parent.reset_mock()
        gene_entry = self.client._extract_gene_of_feature(feature_entry)
        mock_parent.assert_called_with(13, [91])
        self.assertIn(unittest.mock.call(12, [92]), mock_parent.mock_calls)
        self.assertIs(gene_entry, other_feature_entry)

        # Feature is something else - return None
        mock_type.return_value = "chromosome"
        mock_parent.reset_mock()
        gene_entry = self.client._extract_gene_of_feature(feature_entry)
        mock_parent.assert_not_called()
        self.assertIsNone(gene_entry)

    @unittest.mock.patch("pychado.io.gaf.GAFClient.query_child_features")
    @unittest.mock.patch("pychado.io.gaf.GAFClient.query_parent_features")
    @unittest.mock.patch("pychado.io.gaf.GAFClient._extract_feature_type")
    def test_extract_transcript_of_feature(self, mock_type: unittest.mock.Mock, mock_parent: unittest.mock.Mock,
                                           mock_child: unittest.mock.Mock):
        # Tests the function that extracts the transcript associated with a feature from the database
        self.assertIs(mock_type, self.client._extract_feature_type)
        self.assertIs(mock_parent, self.client.query_parent_features)
        self.assertIs(mock_child, self.client.query_child_features)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        other_feature_entry = sequence.Feature(organism_id=11, type_id=300, uniquename="othername", feature_id=13)

        # Feature is a transcript - return input
        mock_type.return_value = "tRNA"
        transcript_entry = self.client._extract_transcript_of_feature(feature_entry)
        mock_type.assert_called_with(feature_entry)
        mock_parent.assert_not_called()
        mock_child.assert_not_called()
        self.assertIs(transcript_entry, feature_entry)

        # Feature is a polypeptide - return parent
        mock_type.return_value = "polypeptide"
        mock_parent.reset_mock()
        mock_child.reset_mock()
        mock_parent.return_value.configure_mock(**{"first.return_value": other_feature_entry})
        transcript_entry = self.client._extract_transcript_of_feature(feature_entry)
        mock_parent.assert_called_with(12, [92])
        mock_child.assert_not_called()
        self.assertIs(transcript_entry, other_feature_entry)

        # Feature is a gene - return child
        mock_type.return_value = "gene"
        mock_parent.reset_mock()
        mock_child.reset_mock()
        mock_child.return_value.configure_mock(**{"first.return_value": other_feature_entry})
        transcript_entry = self.client._extract_transcript_of_feature(feature_entry)
        mock_parent.assert_not_called()
        mock_child.assert_called_with(12, 91)
        self.assertIs(transcript_entry, other_feature_entry)

        # Feature is something else - return None
        mock_type.return_value = "chromosome"
        mock_parent.reset_mock()
        mock_child.reset_mock()
        transcript_entry = self.client._extract_transcript_of_feature(feature_entry)
        mock_parent.assert_not_called()
        mock_child.assert_not_called()
        self.assertIsNone(transcript_entry)

    @unittest.mock.patch("pychado.io.gaf.GAFClient.query_child_features")
    @unittest.mock.patch("pychado.io.gaf.GAFClient._extract_feature_type")
    def test_extract_polypeptide_of_feature(self, mock_type: unittest.mock.Mock, mock_child: unittest.mock.Mock):
        # Tests the function that extracts the polypeptide associated with a feature from the database
        self.assertIs(mock_type, self.client._extract_feature_type)
        self.assertIs(mock_child, self.client.query_child_features)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        other_feature_entry = sequence.Feature(organism_id=11, type_id=300, uniquename="othername", feature_id=13)

        # Feature is a polypeptide - return input
        mock_type.return_value = "polypeptide"
        protein_entry = self.client._extract_polypeptide_of_feature(feature_entry)
        mock_type.assert_called_with(feature_entry)
        mock_child.assert_not_called()
        self.assertIs(protein_entry, feature_entry)

        # Feature is a transcript - return child
        mock_type.return_value = "tRNA"
        mock_child.reset_mock()
        mock_child.return_value.configure_mock(**{"first.return_value": other_feature_entry})
        protein_entry = self.client._extract_polypeptide_of_feature(feature_entry)
        mock_child.assert_called_with(12, 92)
        self.assertIs(protein_entry, other_feature_entry)

        # Feature is a gene - return grandchild
        mock_type.return_value = "gene"
        mock_child.reset_mock()
        protein_entry = self.client._extract_polypeptide_of_feature(feature_entry)
        mock_child.assert_called_with(13, 92)
        self.assertIn(unittest.mock.call(12, 91), mock_child.mock_calls)
        self.assertIs(protein_entry, other_feature_entry)

        # Feature is something else - return None
        mock_type.return_value = "chromosome"
        mock_child.reset_mock()
        protein_entry = self.client._extract_polypeptide_of_feature(feature_entry)
        mock_child.assert_not_called()
        self.assertIsNone(protein_entry)

    def test_convert_evidence_code(self):
        # Tests the function converting a GO evidence code abbreviation into the spelled-out form
        code = self.client._convert_evidence_code("HTP")
        self.assertEqual(code, "Inferred from High Throughput Experiment")
        code = self.client._convert_evidence_code("XYZ")
        self.assertEqual(code, "")

    def test_back_convert_evidence_code(self):
        # Tests the function converting a GO evidence code into its abbreviation
        code = self.client._back_convert_evidence_code("Inferred from High Throughput Experiment")
        self.assertEqual(code, "HTP")
        code = self.client._back_convert_evidence_code("tas")
        self.assertEqual(code, "TAS")
        code = self.client._back_convert_evidence_code("non_existent evidence code")
        self.assertEqual(code, "")

    def test_convert_namespace(self):
        # Tests the function converting a GO namespace into its abbreviation
        abbrev = self.client._convert_namespace("molecular_function")
        self.assertEqual(abbrev, "F")
        abbrev = self.client._convert_namespace("non_existent namespace")
        self.assertEqual(abbrev, "")


if __name__ == '__main__':
    unittest.main(verbosity=2, buffer=True)



import sqlalchemy.orm
import sqlalchemy.sql.functions
from . import base, cv, sequence

# Object-relational mappings for the CHADO Companalysis module


class Analysis(base.PublicBase):
    """Class for the CHADO 'analysis' table"""
    # Columns
    analysis_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    name = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=True)
    desription = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    program = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=False)
    programversion = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=False)
    algorithm = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=True)
    sourcename = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=True)
    sourceversion = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=True)
    sourceuri = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    timeexecuted = sqlalchemy.Column(sqlalchemy.TIMESTAMP, nullable=False,
                                     server_default=sqlalchemy.sql.functions.now())

    # Constraints
    __tablename__ = "analysis"
    __table_args__ = (sqlalchemy.UniqueConstraint(program, programversion, sourcename, name="analysis_c1"),)

    # Initialisation
    def __init__(self, program, programversion, name=None, description=None, algorithm=None, sourcename=None,
                 sourceversion=None, sourceuri=None, timeexecuted=sqlalchemy.sql.functions.now(), analysis_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<companalysis.Analysis(analysis_id={0}, name='{1}', desription='{2}', program='{3}', " \
               "programversion='{4}', algorithm='{5}', sourcename='{6}', sourceversion='{7}, sourceuri='{8}'," \
               "timeexecuted='{9}')>".format(self.analysis_id, self.name, self.desription, self.program,
                                             self.programversion, self.algorithm, self.sourcename, self.sourceversion,
                                             self.sourceuri, self.timeexecuted)


class AnalysisFeature(base.PublicBase):
    """Class for the CHADO 'analysisfeature' table"""
    # Columns
    analysisfeature_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    feature_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        sequence.Feature.feature_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    analysis_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Analysis.analysis_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    rawscore = sqlalchemy.Column(sqlalchemy.FLOAT, nullable=True)
    normscore = sqlalchemy.Column(sqlalchemy.FLOAT, nullable=True)
    significance = sqlalchemy.Column(sqlalchemy.FLOAT, nullable=True)
    identity = sqlalchemy.Column(sqlalchemy.FLOAT, nullable=True)

    # Constraints
    __tablename__ = "analysisfeature"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_id, analysis_id, name="analysisfeature_c1"),
                      sqlalchemy.Index("analysisfeature_idx1", feature_id),
                      sqlalchemy.Index("analysisfeature_idx2", analysis_id))

    # Relationships
    feature = sqlalchemy.orm.relationship(sequence.Feature, foreign_keys=feature_id, backref="analysisfeature_feature")
    analysis = sqlalchemy.orm.relationship(Analysis, foreign_keys=analysis_id, backref="analysisfeature_analysis")

    # Initialisation
    def __init__(self, feature_id, analysis_id, rawscore=None, normscore=None, significance=None, identity=None,
                 analysisfeature_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<companalysis.AnalysisFeature(analysisfeature_id={0}, feature_id={1}, analysis_id={2}, rawscore={3}, " \
               "normscore={4}, significance={5}, identity={6})>".\
            format(self.analysisfeature_id, self.feature_id, self.analysis_id, self.rawscore, self.normscore,
                   self.significance, self.identity)


class AnalysisProp(base.PublicBase):
    """Class for the CHADO 'analysisprop' table"""
    # Columns
    analysisprop_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    analysis_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Analysis.analysis_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    type_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    value = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)

    # Constraints
    __tablename__ = "analysisprop"
    __table_args__ = (sqlalchemy.UniqueConstraint(analysis_id, type_id, value, name="analysisprop_c1"),
                      sqlalchemy.Index("analysisprop_idx1", analysis_id),
                      sqlalchemy.Index("analysisprop_idx2", type_id))

    # Relationships
    analysis = sqlalchemy.orm.relationship(Analysis, foreign_keys=analysis_id, backref="analysisprop_analysis")
    type = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=type_id, backref="analysisprop_type")

    # Initialisation
    def __init__(self, analysis_id, type_id, value=None, analysisprop_id=None):
        for key, val in locals().items():
            if key != self:
                setattr(self, key, val)

    # Representation
    def __repr__(self):
        return "<companalysis.AnalysisProp(analysisprop_id={0}, analysis_id={1}, type_id={2}, value='{3}')>". \
            format(self.analysisprop_id, self.analysis_id, self.type_id, self.value)



import os
import urllib.parse
from typing import Union, List
from Bio import SeqIO, Seq
from . import iobase
from .. import utils
from ..orm import cv, organism, sequence


class FastaImportClient(iobase.ChadoClient):
    """Class for importing genomic data from FASTA files into Chado"""

    def __init__(self, uri: str, verbose=False, test_environment=False):
        """Constructor"""

        # Connect to database
        self.test_environment = test_environment
        if not self.test_environment:
            super().__init__(uri, verbose)

        # Load essentials
        if not self.test_environment:
            self._load_essentials()

    def __del__(self):
        """Destructor - disconnect from database"""
        if not self.test_environment:
            super().__del__()

    def _load_essentials(self) -> None:
        """Loads essential database entries"""
        self._sequence_terms = self._load_terms_from_cv_dict(
            "sequence", ["contig", "supercontig", "chromosome", "region"])
        self._top_level_term = self._load_cvterm("top_level_seq")

    def load(self, filename: str, organism_name: str, sequence_type: str):
        """Import data from a FASTA file into a Chado database"""

        # Load dependencies
        default_organism = self._load_organism(organism_name)
        default_type = self._sequence_terms[sequence_type]

        # Check for file existence
        if not os.path.exists(filename):
            raise iobase.InputFileError("Input file '" + filename + "' does not exist.")

        # Loop over all entries in the FASTA file
        for record in SeqIO.parse(filename, "fasta"):

            # Insert or update entries in the 'feature' table
            feature_entry = self._handle_sequence(record, default_organism, default_type)
            self._mark_as_top_level_sequence(feature_entry)

        # Commit changes
        self.session.commit()

    def _handle_sequence(self, fasta_record: SeqIO.SeqRecord, organism_entry: organism.Organism,
                         default_type_entry: cv.CvTerm) -> sequence.Feature:
        """Inserts or updates an entry in the 'feature' table and returns it"""

        # Check if all dependencies are met. Use default if not.
        sequence_type = self._extract_type(fasta_record)
        if sequence_type:
            if sequence_type not in self._sequence_terms:
                self.printer.print("WARNING: Sequence type '" + sequence_type + "' not present in database")
                type_entry = default_type_entry
            else:
                type_entry = self._sequence_terms[sequence_type]
        else:
            type_entry = default_type_entry

        # Create a feature object, and update the corresponding table
        new_feature_entry = self._create_feature(fasta_record, organism_entry.organism_id, type_entry.cvterm_id)
        feature_entry = self._handle_feature(new_feature_entry, organism_entry.abbreviation)
        return feature_entry

    def _mark_as_top_level_sequence(self, feature_entry: sequence.Feature) -> sequence.FeatureProp:
        """Inserts or updates an entry in the 'featureprop' table and returns it"""
        existing_featureprops = self.query_all(sequence.FeatureProp, feature_id=feature_entry.feature_id)
        new_featureprop_entry = sequence.FeatureProp(feature_id=feature_entry.feature_id,
                                                     type_id=self._top_level_term.cvterm_id, value="true")
        featureprop_entry = self._handle_featureprop(new_featureprop_entry, existing_featureprops,
                                                     self._top_level_term.name, new_featureprop_entry.value,
                                                     feature_entry.uniquename)
        return featureprop_entry

    @staticmethod
    def _create_feature(fasta_record: SeqIO.SeqRecord, organism_id: int, type_id: int) -> sequence.Feature:
        """Creates a feature object from a FASTA record"""
        residues = str(fasta_record.seq)
        return sequence.Feature(organism_id=organism_id, type_id=type_id, uniquename=fasta_record.id,
                                residues=residues, seqlen=len(residues))

    @staticmethod
    def _extract_type(fasta_record: SeqIO.SeqRecord) -> Union[None, str]:
        """Extracts the feature type from a FASTA record"""
        attributes = fasta_record.description.split('|')
        key_value_attributes = dict(attribute.split("=", 1) for attribute in attributes if "=" in attribute)
        for key, value in key_value_attributes.items():
            if key.strip() in ["SO", "sequence_type"]:
                return value.strip()
        return None


class FastaExportClient(iobase.ChadoClient):
    """Class for exporting genomic data from Chado to FASTA files"""

    def __init__(self, uri: str, verbose=False, test_environment=False):
        """Constructor"""

        # Connect to database
        self.test_environment = test_environment
        if not self.test_environment:
            super().__init__(uri, verbose)

        # Load essentials
        if not self.test_environment:
            self._load_essentials()

    def __del__(self):
        """Destructor - disconnect from database"""
        if not self.test_environment:
            super().__del__()

    def _load_essentials(self) -> None:
        """Loads essential database entries"""
        self._sequence_terms = self._load_terms_from_cv_dict(
            "sequence", ["gene", "polypeptide"])
        self._part_of_term = self._load_cvterm_from_cv("part_of", "relationship")
        self._derives_from_term = self._load_cvterm_from_cv("derives_from", "sequence")
        self._top_level_term = self._load_cvterm("top_level_seq")

    def export(self, filename: str, organism_name: str, sequence_type: str, release: str, extract_version=False,
               include_obsolete_features=False):
        """Exports sequences from Chado to a FASTA file"""

        # Load dependencies and features of interest
        organism_entry = self._load_organism(organism_name)
        feature_entries = self._extract_features_by_type(organism_entry, sequence_type)
        srcfeature_entries = self._extract_srcfeatures_by_type(organism_entry, sequence_type)
        records = []

        # Loop over all features of interest
        for feature_entry in feature_entries:

            # Get feature type
            type_entry = self.query_first(cv.CvTerm, cvterm_id=feature_entry.type_id)

            # Create FASTA record
            residues = self._extract_residues_by_type(feature_entry, srcfeature_entries, sequence_type)
            if self._are_residues_valid(residues, sequence_type) and \
                    (include_obsolete_features or not feature_entry.is_obsolete):
                record = self._create_fasta_record(feature_entry, organism_entry, type_entry, residues, release,
                                                   extract_version)
                records.append(record)

        # Write all FASTA records to file
        records.sort(key=self._sort_record_key)
        SeqIO.write(records, filename, "fasta")

    @staticmethod
    def _sort_record_key(record: SeqIO.SeqRecord):
        """Helper function for sorting records"""
        return record.id

    def _create_fasta_record(self, feature_entry: sequence.Feature, organism_entry: organism.Organism,
                             type_entry: cv.CvTerm, residues: str, release: str, extract_version: bool
                             ) -> Union[None, SeqIO.SeqRecord]:
        """Creates a FASTA record"""
        attributes = self._create_fasta_attributes(organism_entry, type_entry, release, extract_version)
        record = SeqIO.SeqRecord(Seq.Seq(residues), id=feature_entry.uniquename, name=feature_entry.uniquename,
                                 description=attributes)
        return record

    def _extract_srcfeatures_by_type(self, organism_entry: organism.Organism, sequence_type: str
                                     ) -> List[sequence.Feature]:
        """Extract features from the database"""
        srcfeature_entries = []
        if sequence_type == "genes":
            srcfeature_entries = self.query_features_by_property_type(organism_entry.organism_id,
                                                                      self._top_level_term.cvterm_id).all()
        return srcfeature_entries

    def _extract_features_by_type(self, organism_entry: organism.Organism, sequence_type: str
                                  ) -> List[sequence.Feature]:
        """Extract features from the database"""
        if sequence_type == "proteins":
            query = self.query_protein_features(organism_entry.organism_id, self._sequence_terms["gene"].cvterm_id,
                                                self._part_of_term.cvterm_id, self._derives_from_term.cvterm_id)
        elif sequence_type == "genes":
            query = self.query_features_by_type(
                organism_entry.organism_id, [self._sequence_terms["gene"].cvterm_id])
        else:
            query = self.query_features_by_property_type(organism_entry.organism_id, self._top_level_term.cvterm_id)
        return query.all()

    def _extract_residues_by_type(self, feature_entry: sequence.Feature, srcfeature_entries: List[sequence.Feature],
                                  sequence_type: str) -> Union[None, str]:
        """Extracts the sequence of nucleotides/amino acids of a feature"""
        if sequence_type == "genes":
            residues = self._extract_nucleotide_sequence(feature_entry, srcfeature_entries)
        else:
            residues = feature_entry.residues
        return residues

    @staticmethod
    def _are_residues_valid(residues: str, sequence_type: str) -> bool:
        """Checks if the sequences of nucleotides/amino acids are composed of valid IUPAC codes"""
        if not residues:
            return False
        if sequence_type == "proteins":
            if residues[0].upper() != "M":
                return False
            if residues[-1].upper() not in Seq.IUPAC.ExtendedIUPACProtein.letters and residues[-1].upper() != "*":
                return False
            for residue in residues[1:-1]:
                if residue.upper() not in Seq.IUPAC.ExtendedIUPACProtein.letters:
                    return False
        else:
            for residue in residues:
                if residue.upper() not in Seq.IUPAC.IUPACAmbiguousDNA.letters:
                    return False
        return True

    def _extract_nucleotide_sequence(self, feature_entry: sequence.Feature, srcfeature_entries: List[sequence.Feature]
                                     ) -> Union[None, str]:
        """Extracts the nucleotide sequence of a (gene) feature"""

        # Get the entry from the 'featureloc' table
        featureloc_entry = self.query_first(sequence.FeatureLoc,
                                            feature_id=feature_entry.feature_id)        # type: sequence.FeatureLoc
        if not featureloc_entry:
            return None

        # Get the entry from the 'feature' table
        matching_entries = utils.filter_objects(
            srcfeature_entries, feature_id=featureloc_entry.srcfeature_id)              # type: List[sequence.Feature]
        if not matching_entries:
            return None

        # Extract the nucleotide sequence
        matching_entry = matching_entries[0]
        if len(matching_entry.residues) >= featureloc_entry.fmax + 1:
            residues = matching_entry.residues[featureloc_entry.fmin:featureloc_entry.fmax].upper()
        else:
            residues = None

        # Compute the complementary sequence, if necessary
        if residues and featureloc_entry.strand < 0:
            sequence_object = Seq.Seq(residues, alphabet=Seq.IUPAC.ambiguous_dna)
            residues = str(sequence_object.reverse_complement())
        return residues

    def _create_fasta_attributes(self, organism_entry: organism.Organism, type_entry: cv.CvTerm, release: str,
                                 extract_version: bool) -> str:
        """Creates a header line for a FASTA sequence with several attributes"""
        attributes_as_list = ["", self._organism_key_value_pair(organism_entry), self._type_key_value_pair(type_entry)]
        if extract_version and organism_entry.version:
            attributes_as_list.append(self._genome_version_key_value_pair(organism_entry))
        if release:
            attributes_as_list.append(self._release_key_value_pair(release))
        attributes_as_string = " | ".join(attributes_as_list).strip()
        return attributes_as_string

    @staticmethod
    def _organism_key_value_pair(organism_entry: organism.Organism):
        """Creates a key-value pair for the FASTA header with the organism name"""
        organism_key = "organism"
        organism_name_as_list = [organism_entry.genus, organism_entry.species]
        if organism_entry.infraspecific_name:
            organism_name_as_list.append(organism_entry.infraspecific_name)
        organism_name_as_string = urllib.parse.quote(" ".join(organism_name_as_list))
        organism_pair = "=".join([organism_key, organism_name_as_string])
        return organism_pair

    @staticmethod
    def _genome_version_key_value_pair(organism_entry: organism.Organism):
        """Creates a key-value pair for the FASTA header with the genome version"""
        version_key = "genome_version"
        version_number = str(organism_entry.version)
        version_pair = "=".join([version_key, version_number])
        return version_pair

    @staticmethod
    def _type_key_value_pair(type_entry: cv.CvTerm):
        """Creates a key-value pair for the FASTA header with the type of the sequence"""
        type_key = "sequence_type"
        type_name = urllib.parse.quote(type_entry.name)
        type_pair = "=".join([type_key, type_name])
        return type_pair

    @staticmethod
    def _release_key_value_pair(release: str):
        """Creates a key-value pair for the FASTA header with the name of the release"""
        release_key = "release"
        release_name = urllib.parse.quote(release)
        release_pair = "=".join([release_key, release_name])
        return release_pair



import pkg_resources
import sqlalchemy.sql.expression
from . import utils


def load_query(specifier: str) -> str:
    """Loads the SQL query for a 'chado extract' command"""
    query = ""
    if specifier == "organisms":
        query = utils.read_text(pkg_resources.resource_filename("pychado", "sql/extract_organisms.sql"))
    elif specifier == "cvterms":
        query = utils.read_text(pkg_resources.resource_filename("pychado", "sql/extract_cvterms.sql"))
    elif specifier == "genedb_products":
        query = utils.read_text(pkg_resources.resource_filename("pychado", "sql/extract_genedb_products.sql"))
    elif specifier == "stats":
        query = utils.read_text(pkg_resources.resource_filename("pychado", "sql/extract_stats.sql"))
    return query


def set_query_conditions(query: str, **kwargs) -> sqlalchemy.sql.expression.TextClause:
    """Replaces placeholder in a query and binds parameters"""

    # Set/remove conditions in WHERE clause
    if "organism" in kwargs.keys():
        query = set_organism_condition(query, kwargs["organism"])
    if "database" in kwargs.keys():
        query = set_database_condition(query, kwargs["database"])
    if "vocabulary" in kwargs.keys():
        query = set_vocabulary_condition(query, kwargs["vocabulary"])

    # Bind parameters
    final_query = bind_parameters(query, **kwargs)
    return final_query


def bind_parameters(query: str, **kwargs) -> sqlalchemy.sql.expression.TextClause:
    """Binds parameters to a given query"""
    text_query = sqlalchemy.text(query)
    for key, value in kwargs.items():
        if value:
            text_query = text_query.bindparams(sqlalchemy.bindparam(key, value=value))
    return text_query


def set_organism_condition(query: str, organism: str) -> str:
    """Replaces a placeholder in a query with a condition restricting results to a certain organism"""
    if not organism:
        modified_query = query.replace(":ORGANISM_CONDITION", "TRUE")
    else:
        modified_query = query.replace(":ORGANISM_CONDITION", "abbreviation = :organism")
    return modified_query


def set_database_condition(query: str, database: str) -> str:
    """Replaces a placeholder in a query with a condition restricting results to a certain database"""
    if not database:
        modified_query = query.replace(":DB_CONDITION", "TRUE")
    else:
        modified_query = query.replace(":DB_CONDITION", "db.name = :database")
    return modified_query


def set_vocabulary_condition(query: str, vocabulary: str) -> str:
    """Replaces a placeholder in a query with a condition restricting results to a certain vocabulary"""
    if not vocabulary:
        modified_query = query.replace(":CV_CONDITION", "TRUE")
    else:
        modified_query = query.replace(":CV_CONDITION", "cv.name = :vocabulary")
    return modified_query



import sqlalchemy.orm
from . import base

# Object-relational mappings for the CHADO General module


class Db(base.PublicBase):
    """Class for the CHADO 'db' table"""
    # Columns
    db_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    name = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=False)
    description = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=True)
    urlprefix = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=True)
    url = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=True)

    # Constraints
    __tablename__ = "db"
    __table_args__ = (sqlalchemy.UniqueConstraint(name, name="db_c1"), )

    # Initialisation
    def __init__(self, name, description=None, urlprefix=None, url=None, db_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<general.Db(db_id={0}, name='{1}', description='{2}', urlprefix='{3}', url='{4}')>"\
            .format(self.db_id, self.name, self.description, self.urlprefix, self.url)


class DbxRef(base.PublicBase):
    """Class for the CHADO 'dbxref' table"""
    # Columns
    dbxref_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    db_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Db.db_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    accession = sqlalchemy.Column(sqlalchemy.VARCHAR(1024), nullable=False)
    version = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=False, server_default=sqlalchemy.text("''"))
    description = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)

    # Constraints
    __tablename__ = "dbxref"
    __table_args__ = (sqlalchemy.UniqueConstraint(db_id, accession, version, name="dbxref_c1"),
                      sqlalchemy.Index("dbxref_idx1", db_id),
                      sqlalchemy.Index("dbxref_idx2", accession),
                      sqlalchemy.Index("dbxref_idx3", version))

    # Relationships
    db = sqlalchemy.orm.relationship(Db, backref="dbxref_db")

    # Initialisation
    def __init__(self, db_id, accession, version="", description=None, dbxref_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<general.DbxRef(dbxref_id={0}, db_id={1}, accession='{2}', version='{3}', description='{4}')>"\
            .format(self.dbxref_id, self.db_id, self.accession, self.version, self.description)



import shutil
import pkg_resources
import subprocess
import urllib.parse
import json
import getpass
import sqlalchemy.engine
import sqlalchemy.sql
import sqlalchemy_utils
from . import utils


def set_default_parameters() -> None:
    """Sets the default connection parameters"""
    parameters = {}
    print("Please set the connection parameters for your default database.")
    parameters["host"] = input("host: ")
    parameters["port"] = input("port: ")
    parameters["database"] = input("database: ")
    parameters["user"] = input("username: ")
    parameters["password"] = getpass.getpass(prompt="password: ")
    utils.dump_yaml(default_configuration_file(), parameters)
    print("Your default connection parameters have been changed.")


def reset_default_parameters() -> None:
    """Resets the default connection parameters to factory settings"""
    shutil.copyfile(factory_settings_configuration_file(), default_configuration_file())
    print("Your default connection parameters have been reset.")


def generate_uri(connection_details: dict) -> str:
    """Creates a connection URI"""
    uri_as_list = ["postgresql://"]
    if "user" in connection_details and connection_details["user"] is not None:
        uri_as_list.append(urllib.parse.quote(connection_details["user"]))
        if "password" in connection_details and connection_details["password"] is not None:
            uri_as_list.append(":" + urllib.parse.quote(connection_details["password"]))
        uri_as_list.append("@")
    if "host" in connection_details and connection_details["host"] is not None:
        uri_as_list.append(urllib.parse.quote(connection_details["host"]))
    if "port" in connection_details and connection_details["port"] is not None:
        uri_as_list.append(":" + connection_details["port"])
    if "database" in connection_details and connection_details["database"] is not None:
        uri_as_list.append("/" + urllib.parse.quote(connection_details["database"]))
    return "".join(uri_as_list)


def default_schema_url() -> str:
    """Returns the URL of the default schema"""
    yaml_file = pkg_resources.resource_filename("pychado", "data/gmodSchema.yml")
    default_schema = utils.parse_yaml(yaml_file)
    return default_schema["url"].replace("<VERSION>", default_schema["version"])


def default_configuration_file() -> str:
    """Returns the name of the default configuration file"""
    return pkg_resources.resource_filename("pychado", "data/defaultDatabase.yml")


def factory_settings_configuration_file() -> str:
    """Returns the name of the configuration file with factory settings"""
    return pkg_resources.resource_filename("pychado", "data/factorySettings.yml")


def open_connection(uri: str) -> sqlalchemy.engine.base.Connection:
    """Establishes a database connection"""
    engine = sqlalchemy.create_engine(uri)
    connection = engine.connect()
    return connection


def close_connection(connection: sqlalchemy.engine.base.Connection) -> None:
    """Closes a database connection"""
    connection.close()
    connection.engine.dispose()


def exists(uri: str) -> bool:
    """Checks if a database exists"""
    return sqlalchemy_utils.database_exists(uri)


def random_database_uri(connection_parameters: dict) -> str:
    """Generates a random database name and makes sure the name is not yet in use"""
    parameters = connection_parameters.copy()
    while True:
        parameters["database"] = utils.random_string(10)
        uri = generate_uri(parameters)
        if not exists(uri):
            break
    return uri


def create_database(uri: str) -> None:
    """Creates a database"""
    sqlalchemy_utils.create_database(uri)
    print("Database has been created.")


def drop_database(uri: str, force=False) -> None:
    """Deletes a database"""
    if not force:
        dbname = uri.split("/")[-1]
        if input("Are you sure you want to drop the database '" + dbname + "' [y/n]?\n").lower() != "y":
            print("Database is conserved.")
            return
    sqlalchemy_utils.drop_database(uri)
    print("Database has been deleted.")


def dump_database(uri: str, archive: str) -> None:
    """Dumps a database into an archive file"""
    command = ["pg_dump", "-f", archive, "--format=custom", uri]
    subprocess.run(command)
    print("Database has been dumped.")


def restore_database(uri: str, archive: str) -> None:
    """Restores a database from an archive file"""
    command = ["pg_restore", "--clean", "--if-exists", "--no-owner", "--no-privileges", "--format=custom",
               "-d", uri, archive]
    subprocess.run(command)
    print("Database has been restored.")


def setup_database(uri: str, schema_file: str) -> None:
    """Sets up a database according to a given schema"""
    command = ["psql", "-q", "-f", schema_file, uri]
    subprocess.run(command)
    print("Database schema has been set up.")


def connect_to_database(uri: str) -> None:
    """Connects to a database for an interactive session"""
    print("Establishing connection to database...")
    command = ["psql", uri]
    subprocess.run(command)
    print("Connection to database closed.")


def query_and_print(uri: str, query, filename: str, file_format: str, header: bool, delimiter: str) -> None:
    """Opens a database connection, runs a single query, and prints the query result"""
    conn = open_connection(uri)
    result = run_query(conn, query)
    print_query_result(result, filename, file_format, header, delimiter)
    result.close()
    close_connection(conn)


def run_query(conn: sqlalchemy.engine.Connection, query) -> sqlalchemy.engine.ResultProxy:
    """Runs a single query against a database"""
    if isinstance(query, sqlalchemy.sql.expression.TextClause) or isinstance(query, sqlalchemy.sql.Select):
        result = conn.execute(query)
    else:
        result = conn.execute(sqlalchemy.text(query))
    return result


def print_query_result(result: sqlalchemy.engine.ResultProxy, filename: str, file_format: str, header: bool,
                       delimiter: str) -> None:
    """Exports a table resulting from a database query"""
    file_handle = utils.open_file_write(filename)
    if file_format == "csv":
        print_query_result_csv(result, file_handle, header, delimiter)
    elif file_format == "json":
        print_query_result_json(result, file_handle)
    else:
        print("Unknown file format: '" + file_format + "'")
    utils.close(file_handle)


def print_query_result_csv(result: sqlalchemy.engine.ResultProxy, file_handle, header: bool, delimiter: str) -> None:
    """Exports a table resulting from a database query as CSV"""
    keys = result.keys()
    if header:
        file_handle.write(utils.list_to_string(keys, delimiter) + "\n")
    while True:
        row = result.fetchone()
        if not row:
            break
        file_handle.write(utils.list_to_string(row, delimiter) + "\n")


def print_query_result_json(result: sqlalchemy.engine.ResultProxy, file_handle) -> None:
    """Exports a table resulting from a database query as JSON"""
    keys = result.keys()
    json_obj = []
    while True:
        row = result.fetchone()
        if not row:
            break
        json_dict = dict(zip(keys, row))
        json_obj.append(json_dict)
    json.dump(json_obj, file_handle, indent=4, sort_keys=True)
    file_handle.write("\n")



import sqlalchemy.ext.declarative

# Supplies the base for declarative mappings
PublicBase = sqlalchemy.ext.declarative.declarative_base(metadata=sqlalchemy.schema.MetaData(schema='public'))
AuditBase = sqlalchemy.ext.declarative.declarative_base(metadata=sqlalchemy.schema.MetaData(schema='audit'))

# Define data types
BIGINT = sqlalchemy.BIGINT().with_variant(sqlalchemy.INTEGER(), 'sqlite')
operation_type = sqlalchemy.Enum('INSERT', 'UPDATE', 'DELETE', name='operation_type')



import sys
import os
import unittest
import subprocess
import string
import io
import filecmp
import tempfile
from contextlib import redirect_stdout
from .. import utils


class TestUtils(unittest.TestCase):
    """Class for testing utilities"""

    modules_dir = os.path.dirname(os.path.abspath(utils.__file__))
    data_dir = os.path.join(modules_dir, 'tests', 'data')

    def test_write_and_read(self):
        # open_file_write() and open_file_read() should do the right thing whether gzipped or not
        for filename in ['utils.tmp', 'utils.tmp.gz']:
            # Test write
            f = utils.open_file_write(filename)
            for i in range(3):
                print(i, file=f)
            utils.close(f)
            # Test read
            counter = 0
            f = utils.open_file_read(filename)
            for line in f:
                self.assertEqual(counter, int(line.strip()))
                counter += 1
            utils.close(f)
            # Remove files
            os.unlink(filename)
        # Test streams stdin/stdout
        f = utils.open_file_read('')
        self.assertEqual(sys.stdin, f)
        f = utils.open_file_write('')
        self.assertEqual(sys.stdout, f)

    def test_write_read_text(self):
        # tests reading and writing text from/to file
        text = utils.random_string(100)
        filename = tempfile.mkstemp()[1]
        utils.write_text(filename, text)
        self.assertTrue(os.path.exists(os.path.abspath(filename)))
        read_text = utils.read_text(filename)
        self.assertEqual(text, read_text)
        os.remove(filename)
        self.assertFalse(os.path.exists(os.path.abspath(filename)))

    def test_raise_exception(self):
        # open_file_write() and open_file_read() should raise an exception if opening fails
        with self.assertRaises(FileNotFoundError):
            utils.open_file_read('this_file_is_not_here_so_throw_error')
        with self.assertRaises(FileNotFoundError):
            utils.open_file_read('this_file_is_not_here_so_throw_error.gz')
        with self.assertRaises(subprocess.CalledProcessError):
            utils.open_file_read(os.path.join(self.data_dir, 'utils_test_not_really_zipped.gz'))
        with self.assertRaises(FileNotFoundError):
            utils.open_file_write(os.path.join('not_a_directory', 'this_file_is_not_here_so_throw_error'))
        with self.assertRaises(FileNotFoundError):
            utils.open_file_write(os.path.join('not_a_directory', 'this_file_is_not_here_so_throw_error.gz'))

    def test_write_csv(self):
        # checks that data are correctly written into a CSV file
        data = [["name", "clade", "legs", "extinct"],
                ["leech", "annelid", "0", "f"],
                ["human", "mammal", "2", "f"],
                ["diplodocus", "reptile", "4", "t"],
                ["bumblebee", "insect", "6", "f"]]
        filename = tempfile.mkstemp()[1]
        self.assertTrue(os.path.exists(os.path.abspath(filename)))
        utils.write_csv(filename, ";", data)
        self.assertTrue(filecmp.cmp(filename, os.path.join(self.data_dir, "dbutils_species_table.csv")))
        os.remove(filename)
        self.assertFalse(os.path.exists(os.path.abspath(filename)))

    def test_yaml_parser(self):
        # checks if a yaml file is parsed correctly
        filename = os.path.join(self.data_dir, "utils_yaml_example.yml")
        content = utils.parse_yaml(filename)
        self.assertIn("institute", content)
        self.assertEqual(content["institute"], "Sanger")
        self.assertIn("founded", content)
        self.assertEqual(content["founded"], "1993")
        self.assertIn("faculties", content)
        self.assertIn("parasites and microbes", content["faculties"])
        self.assertNotIn("zebrafish genetics", content["faculties"])

    def test_list_to_string(self):
        # checks if a list is correctly concatenated
        test_list = [1.123, None, 'hello', True, 'A', 8, False]
        test_string = utils.list_to_string(test_list, "_")
        self.assertEqual(test_string, "1.123__hello_t_A_8_f")
        test_string_with_prefix = utils.list_to_string(test_list, "\t", "A")
        self.assertEqual(test_string_with_prefix, "A.1.123\tA.\tA.hello\tA.t\tA.A\tA.8\tA.f")

    def test_filter_objects(self):
        # checks if a function correctly filters objects in a list according to keyword arguments
        john = utils.EmptyObject(name="John", age=42, sex="m")
        mike = utils.EmptyObject(name="Mike", age=23, sex="m")
        persons = [john, mike]
        filtered_persons = utils.filter_objects(persons, name="Mike", sex="m")
        self.assertEqual(len(filtered_persons), 1)
        self.assertEqual(filtered_persons[0], mike)
        with self.assertRaises(AttributeError):
            utils.filter_objects(persons, heads=2)

    def test_copy_attributes(self):
        # checks if a function correctly copies attributes from one object to another
        john = utils.EmptyObject(name="John", age=42, sex="m")
        mike = utils.EmptyObject(name="Mike", age=23, sex="m")
        updated = utils.copy_attribute(john, mike, "age")
        self.assertTrue(updated)
        self.assertEqual(getattr(john, "age"), 23)
        updated = utils.copy_attribute(john, mike, "sex")
        self.assertFalse(updated)
        updated = utils.copy_attribute("abc", "def", "xyz")
        self.assertFalse(updated)
        updated = utils.copy_attribute("abc", 567, "xyz")
        self.assertFalse(updated)

    def test_list_to_dict(self):
        # checks if a function correctly converts a list into a dictionary
        john = utils.EmptyObject(name="John")
        mike = utils.EmptyObject(name="Mike")
        persons = [john, mike]
        persons_dict = utils.list_to_dict(persons, "name")
        self.assertEqual(len(persons_dict), 2)
        self.assertIn("John", persons_dict)
        self.assertEqual(persons_dict["John"], john)
        with self.assertRaises(AttributeError):
            utils.list_to_dict(persons, "age")

    def test_random_string(self):
        # tests if a function generates a random string of lowercase letters
        string1 = utils.random_string(8)
        self.assertEqual(len(string1), 8)
        for letter in string1:
            self.assertIn(letter, string.ascii_lowercase)
        string2 = utils.random_string(8)
        self.assertNotEqual(string1, string2)

    def test_random_integer(self):
        # tests if a function generates a random integer
        int1 = utils.random_integer(10000)
        self.assertLessEqual(int1, 10000)
        self.assertGreaterEqual(int1, 0)
        int2 = utils.random_integer(10000)
        self.assertNotEqual(int1, int2)

    def test_random_float(self):
        # tests if a function generates a random float number between 0 and 1
        float1 = utils.random_float()
        self.assertGreaterEqual(float1, 0.0)
        self.assertLessEqual(float1, 1.0)
        float2 = utils.random_float()
        self.assertFalse(abs(float1 - float2) < 0.00001)

    def test_current_date(self):
        # checks if a function returns the current date in the correct format
        date = utils.current_date()
        self.assertEqual(len(date), 8)
        self.assertEqual(int(date[0]), 2)
        self.assertTrue(int(date[4:5]) <= 12)
        self.assertTrue(int(date[6:7]) <= 31)

    def test_verbose_printer(self):
        # tests the verpose printer
        printer = utils.VerbosePrinter(True, "-")
        f = io.StringIO()
        with redirect_stdout(f):
            printer.print(["AAA", "BBB"])
            printed = f.getvalue()
        self.assertEqual(printed, "AAA-BBB\n")


if __name__ == '__main__':
    unittest.main(verbosity=2, buffer=True)
